! Time-stamp: <e0500.fsl.noaa.gov:/home/jacobson/TM5/cy2/proj/chem/enkf/inv/src/enkf_tools.F90: 01 May 2007 (Tue) 23:39:43 GMT>

!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module enkf_tools

  !  These are declarations and routines needed for the inverse run
  !  of the ensemble Kalman filter.
  !
  !  Table of Contents:
  !
  ! module enkf_tools
  !   subroutine Filter_Init(status)
  !   end subroutine Filter_Init
  !   subroutine CreateEnsemble
  !     subroutine MakePriorParams(nstate)
  !     end subroutine MakePriorParams
  !     subroutine MakePriorUncertainty(status)
  !     end subroutine MakePriorUncertainty
  !     subroutine MakeEnsemble(nstate,status)
  !     end subroutine MakeEnsemble
  !     subroutine ReadState(file_name,status)
  !     end subroutine ReadState
  !   end subroutine CreateEnsemble
  !   subroutine Vectorize(obsonly)
  !     subroutine GatherObs
  !     end subroutine GatherObs
  !     subroutine GatherState
  !     end subroutine GatherState
  !   end subroutine Vectorize
  !   subroutine unVectorize(obsonly)
  !     subroutine ScatterState
  !     end subroutine ScatterState
  !   end subroutine unVectorize
  !   subroutine SampleState
  !   end subroutine SampleState
  !   subroutine AdvanceBackground
  !   end subroutine AdvanceBackground
  !   subroutine Invert
  !     subroutine make_K
  !     end subroutine make_K
  !     subroutine localize
  !     end subroutine localize
  !     real function correlate(x,y,n)
  !     end function correlate
  !     subroutine solve_enkf
  !     end subroutine solve_enkf
  !     subroutine save_invert(file_name,msg,flag)
  !     end subroutine save_invert
  !   end subroutine Invert
  !   subroutine checkseed
  !   end subroutine checkseed
  ! end module enkf_tools
  
  use GO,                  only: gol, goErr, goPr, goLabel
  use GO,                  only: TrcFile, Init, Done, ReadRc
  use GO,                  only: TDate, NewDate, IncrDate, operator(+), operator(-), rTotal
  use toolbox,             only: escape_tm
  use global_data,         only: rcfile
  use ParTools,            only: myid, root, maxtrac_per_pe, Par_Barrier, ntracetloc
  use ParTools,            only: proc_tracer, tracer_loc, tracer_active, npes,ntracet_ar
  use chem_param,          only: ntracet
  use user_output_forecast,only: n_fcast, concentration, site_categories, sites, global_R_scaling, obs_rejection_threshold,n_sites_active
  use emission_co2_all,    only: pseudo
  use random
  use enkf_common,         only: nlag, timestep, nparams_in_state, area_1x1, regtype, area_state
  use enkf_common,         only: init_scalefac, nreg_land, npar_land, nreg_ocn, npar_ocn, fluxregions
  use enkf_common,         only: state_X, state_Q, Init_Parameters, savestate_filename
  use ParTools,            only: nmembers, nmembers_per_pe, nmembersloc

  implicit none
  ! --- const -------------------------

  character(len=*), parameter  ::  mname = 'enkf_tools'

  ! --- var ---------------------------

  logical                          :: kfdebug=.false.
  logical                          :: ifirst=.false.

  !WP! vectors and matrices needed for inversion algebra

  real,dimension(:,:),allocatable                         :: R             !  model-data mismatch
  real,dimension(:)  ,allocatable                         :: K, corr, alpha       !  Kalman gain, alpha scalar  
  real,dimension(:)  ,allocatable,target                  :: meanstate     !  meanstate as one long vector
  real,dimension(:,:),allocatable,target                  :: dF,dX ! non-normalized deviations of each obs and flux from mean

  !WP! Local variables that are useful

  integer, dimension(6)            :: idate_filter  ! internal time of filter assimilation window
  integer                          :: itau_filter   ! 
  integer,dimension(2)           :: sharedseed    ! holds seed for random number generator
  integer,dimension(2)           :: savedseed     ! holds seed for random number generator
  type(TDate)                    :: tread1, tread2, tdyn, tend
  type(TrcFile)        :: rcF

  character(len=256) :: savedir ! save directory
  character(len=256) :: prev_savestate_filename  ! filename for previous savestate file.

contains

  subroutine Filter_Init(status)
    ! Filter_Init sets up the configuration of the state, reads observations for
    ! the new step, and seeds the random number generator.
    !
    ! call to subroutines:
    !
    ! random_seeds
    ! read_sites
    ! read_co2_obs
    ! CalcAreaState
    ! ReadRegions
    !
    !

    use file_hdf,                only: THdfFile, Init, Done, ReadAttribute
    use datetime,                only: tau2date
    use Dims,                    only: itaui, istart
    use user_output_forecast,    only: read_sites,read_co2_obs,ObsListFFilename
    use ParTools,                only: myid, root, Par_Barrier
#ifdef MPI
    use mpi_const,               only: com_trac, mpi_integer,ierr, mpi_comm_world
#endif
    implicit none
    ! --- in/out -----------------------------

    integer, intent(out)      ::  status
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Filter_Init'

    ! --- local -------------------------------
    integer              :: k,n
    integer,dimension(8) :: value
    character(len=10)    :: dummydate,dummytime,dummyzone, random_method
    type(THdfFile)    ::  hdf

    !WP! Get general settings: pseudo data or not

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'create.pseudo', pseudo ,status ,default=.false.)
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'random.seed.method', random_method ,status ,default='fixed')
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'savedir', savedir ,status)
    IF_NOTOK_RETURN(status=1)
    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)

    ! By assumption, the savestate.hdf file from the previous run is sitting in the save directory.
    ! This is handled by run-tm5.jb.
    prev_savestate_filename = trim(savedir)//'/'//trim(savestate_filename)

    !WP! If we've just started or we're running pseudo data then pretend we have no
    !WP! flux info from previous steps

    if(istart==6.or.istart.eq.1.or.pseudo) ifirst=.true.

    itau_filter=itaui  ! filter starts at start time of run
    call tau2date(itau_filter,idate_filter)

    !WP! Initialize random numbers on all processors. If this is the start of an
    !WP! assimilation (ifirst) then we get our seed from the system clock, or hardcode
    !WP! it. Otherwise, we read the seed from a textfile. Special case of
    !WP! random.seed.method=='repeat' also causes reseeding of filter from rc-file
    !WP! numbers 

    if(myid==root) then

       if(ifirst.or.trim(random_method)=='repeat') then  ! create seed

          call random_seed(size=k)   ! get size of random number seed

          if(k.ne.2) call escape_tm('FilterInit: more than two random seeds requested, is this fortran?)')

          select case(trim(random_method)) 

          case('clock')

             ! date and time as seed

             call date_and_time(dummydate,dummytime,dummyzone,value)
             sharedseed(1)=value(7)  !seconds of the minute
             sharedseed(2)=value(8)  !milliseconds of the second

          case('fixed')

             ! fixed initial seed, note: only for very first step of filter!

             sharedseed(1)=4357
             sharedseed(2)=1100

          case('read','repeat')

             ! read initial seed, note: only for very first step of filter for 'read'!
             ! read initial seed, note: same seed for all steps of filter for 'repeat'!

             call Init( rcF, rcfile, status )        
             IF_NOTOK_RETURN(status=1)
             call ReadRc( rcF, 'random.seed.values', sharedseed, status )
             IF_NOTOK_RETURN(status=1)
             call Done( rcF , status)
             IF_NOTOK_RETURN(status=1)

          case default
             call escape_tm('FilterInit: Unrecognized value of random_method from Rc-file, exiting')
          end select

       else  ! read seed from previous run 

          call Init(hdf,prev_savestate_filename,'read',status)
          IF_NOTOK_RETURN(status=1)
          call ReadAttribute( hdf, 'random_seed', sharedseed, status )
          IF_NOTOK_RETURN(status=1)
          call Done(hdf,status)
          IF_NOTOK_RETURN(status=1)

       endif  ! ifirst

    endif  ! root

    call Par_Barrier

#ifdef MPI
    call mpi_bcast(sharedseed,2,mpi_integer,root,mpi_comm_world,ierr)  ! bcast shared seed from root
#endif

    call random_seed(put=sharedseed(1:2))   ! put sharedseed in seed on each PE

    write (gol,*)'On PE0 random seed initialized as ',sharedseed ; call goPr

    call checkseed ! should be synched

    savedseed=sharedseed  !WP! make equal so we can re-create the random numbers when doing
    !WP! AdvanceBackground

    call Par_Barrier

    call Init_Parameters(status)
    IF_ERROR_RETURN(status=1)

    call read_sites(status)
    IF_ERROR_RETURN(status=1)

    ObsListFFilename='obs_forecast.list' ! list of observations used in optimization
    call read_co2_obs(status) 
    IF_ERROR_RETURN(status=1)

    ! ok
    status = 0

  end subroutine  Filter_Init


  subroutine CreateEnsemble
    !
    ! This subroutine handles all necessary I/O and processes to make prior
    ! fluxes and covariances and/or read those from previous cycles
    !
    use datetime,         only: tstamp
    use Dims,             only: kmain
    use ParTools,         only: Par_Barrier
    implicit none

    character(len=*), parameter ::  rname = mname//'/CreateEnsemble'
    integer :: n
    integer             :: status

    if(ifirst) then   ! on initialization every time step gets a bottom-up prior flux
       do n=1,nlag    ! note that the order of the next 4 calls matters!!!
          call MakePriorUncertainty(status)  ! all PE's make Q-matrix
          if(status/=0) call escape_tm('Stop in '//rname)
          call MakeEnsemble(n,status)       ! create Gaussian parameter ensemble from Q around mean of 0.0 
          if(status/=0) call escape_tm('Stop in '//rname)
          call MakePriorParams(n)    ! every PE gets a first guess
       enddo
       call Par_Barrier
       if (myid==root) call tstamp(kmain,itau_filter,'Done making prior ensemble fields')
    else              ! if not first step, read results from previous runs
       call Init( rcF, rcfile, status )           ! get name of savefile
       if(status/=0) call escape_tm('Stop in '//rname)

       write (gol,*)'EnKF inverse:  Reading previous state from ',prev_savestate_filename; call goPr

       call ReadState(prev_savestate_filename,status)              ! read optimized states from previous cycles
       if(status/=0) call escape_tm('Stop in '//rname)
       call Done( rcF , status)
       if(status/=0) call escape_tm('Stop in '//rname)
       call MakePriorUncertainty(status)          ! all PE's make Q-matrix
       if(status/=0) call escape_tm('Stop in '//rname)
       call MakeEnsemble(nlag,status)            ! make Gaussian distr around 0.0 for new state
       if(status/=0) call escape_tm('Stop in '//rname)
       call DynamicalModel                ! propagate mean state to make prior for new time step
       if(myid==root) call tstamp(kmain,itau_filter,'Done propagating state and covariance')
    endif


  contains
    
    subroutine MakePriorParams(nstate)
      !
      ! Subroutine sets prior values for parameters to be estimated
      !
      implicit none
      !_____________________IO_________________________________________
      integer,intent(in)       :: nstate

      !_____________________locals_____________________________________

      state_X(nstate)%xpc=1.0  ! assume scaling factor on fluxes of 1

    end subroutine MakePriorParams

    subroutine MakePriorUncertainty(status)
      !
      ! Subroutine fills the covariance matrix for a specific time step. This can be
      ! done by just adding variance on the diagonals, or with more complicated
      ! structures.
      use ParTools,             only: myid, root
      use emission_co2_ocean,   only: covariance_co2_compute_ocean
      use emission_co2_bio,     only: covariance_co2_compute_bio
      use file_hdf,             only: THdfFile, TSds, Init, Done, WriteData

      implicit none

      ! i/o
      integer,intent(out) :: status

      !local
      character(len=*), parameter ::  rname = mname//'/MakePriorUncertainty'
      integer :: parm_start,parm_end
      type(THdfFile)      :: hdf
      type(TSds)		:: sds

      state_Q%priorcov(:,:)=0.0  ! set to zero before filling

      parm_start = 1
      parm_end = npar_land*nreg_land
      call covariance_co2_compute_bio(state_Q%priorcov,parm_start,parm_end)

      parm_start = npar_land*nreg_land + 1
      parm_end = parm_start + npar_ocn*nreg_ocn - 1
      call covariance_co2_compute_ocean(state_Q%priorcov,parm_start,parm_end)

      ! Undefined regions
      state_Q%priorcov(nparams_in_state,nparams_in_state) = 1e-10

      if(myid==root) then
         call Init( hdf, "covariance_confirm.hdf", 'create', status )
         IF_ERROR_RETURN(status=1)
         call Init( sds, hdf, "cov", shape(state_Q%priorcov), 'float32', status)
         IF_ERROR_RETURN(status=1)
         call WriteData(sds,state_Q%priorcov,status)
         IF_ERROR_RETURN(status=1)
         call Done (sds,status)
         IF_ERROR_RETURN(status=1)
         call Done (hdf,status)
         IF_ERROR_RETURN(status=1)
      endif

      status = 0

    end subroutine MakePriorUncertainty


    subroutine MakeEnsemble(nstate,status)
      ! From state_Q%priorcov create ensemble members 
      ! from unconditional realizations using the Cholesky decomposition 

      use random,  only: random_normal
      use io_hdf,      only: io_write
      use io_hdf,      only: DFNT_INT32, DFNT_CHAR, DFNT_FLOAT64, DFACC_CREATE, DFACC_WRITE
#ifdef MPI
      use mpi_const, only: com_trac, my_real, ierr,myid,root
#endif
      implicit none
      !i/o
      integer,intent(in) :: nstate
      integer,intent(inout) :: status
      !local
      character(len=*), parameter ::  rname = mname//'/MakeEnsemble'
      integer :: n,m,istat,tag
      real,dimension(nparams_in_state,nparams_in_state) :: C
      real,dimension(nparams_in_state)                   :: randomnumber
      real,dimension(:),pointer                           :: deviation
      integer           :: sfsnatt, sfscatt, io, sfstart
      integer           :: sfcreate, sfscompress, sds_id, sfwdata
      integer           :: sfendacc, sfend, dimid3, dimid0, dimid1
      integer           :: dimid2, sfdimid, sfsdmname
      integer           :: region, ind, imr,jmr, lmr




      !WP! Cholesky decomposition
      C=state_Q%priorcov

      call DPOTRF( 'L' , nparams_in_state, C, nparams_in_state, status ) ! Q data is in lower

      if(status.ne.0) then
         write (gol,*)'Failed Cholesky decomposition',status ; call goPr
         IF_NOTOK_RETURN(status=1)
      endif


      do m=1,nparams_in_state-1
         C(m,m+1:nparams_in_state)=0.0  ! zero out Q data, sqrt(q) is left in upper
      enddo


      do n=1,nmembers
         do m=1,nparams_in_state
            randomnumber(m)=random_normal()  ! random numbers are same on each PE
         enddo

         if(tracer_active(n)) state_X(nstate)%pdX(:,tracer_loc(n)) = matmul(C,randomnumber) !unconditional realization

      enddo

      if(myid==root) state_X(nstate)%pdX(:,1)=0.0 ! mean state has 0.0 deviation

!      if(myid==root) print *,"DEBUG root, pdX(:,3):",state_X(nstate)%pdX(:,3)

      !ok
      status=0

    end subroutine MakeEnsemble

    subroutine ReadState(file_name,status)
      !
      ! Subroutine reads in state from the previous step, and places proper fluxes
      ! into the current state vector.
      !
      use io_hdf
      use GO_string

      implicit none
      ! in/out
      character(len=*), intent(in)    :: file_name 
      integer,intent(out)           :: status

      ! local
      character(len=*), parameter ::  rname = mname//'/ReadState'
      real,dimension(:),pointer       :: ldx,xpc
      integer                         :: io, sfstart, ifail, sds_id
      integer                         :: n_datasets, n_file_attrs
      integer                         :: istat, attributes, num_type
      integer                         :: sffinfo, sfselect, sfginfo
      integer                         :: sfend, sfrnatt, sfrcatt, sffattr
      integer                         ::  i,j,l,n,m,ind,mem
      character(len=80)               :: msg_file
      character(len=2)                :: mem_id
      integer                         :: offsetn=0

      io = sfstart(trim(file_name), DFACC_READ)
      if ( io == FAIL ) then
         write (gol,*)'sfstart failed on file: ',file_name; call goPr
         status=1
         IF_NOTOK_RETURN(status=1)
      endif


      if(myid>0) offsetn=sum(ntracet_ar(0:myid-1))

      allocate(dX(nparams_in_state,nmembers))  ! use dX to read and transfer non-normalized deviations

      do mem=1,nmembersloc

         do n=2,nlag                ! skip n=1 since that dropped out of state

            xpc => state_X(n-1)%xpc ! pointer ensures shift in state position

            call io_read1d_32g(io,nparams_in_state,xpc,'xac_'//goNum2Str(n,'(i2.2)'),status)
            IF_NOTOK_RETURN(status=1)
            call io_read2d_32g(io,nparams_in_state,nmembers,dX,'adX_'//goNum2Str(n,'(i2.2)'),status)
            IF_NOTOK_RETURN(status=1)
            
!            if(myid .eq. root) print *, '[DEBUG] read xac_'//goNum2Str(n,'(i2.2)')//' into xpc_'//goNum2Str(n-1,'(i2.2)')//':',state_X(n-1)%xpc

            state_X(n-1)%pdX(:,mem) = dX(:,offsetn+mem)

            nullify(xpc)

         enddo

      enddo

      deallocate(dX)

      status=sfend(io)
      !write (gol,*)'ReadState: sfend returns ',status; call goPr
      IF_NOTOK_RETURN(status=1)

      status=0  ! all OK

    end subroutine ReadState

    subroutine DynamicalModel
      use enkf_common
      use io_hdf,      only : io_write
      use io_hdf,      only : DFNT_INT32, DFNT_CHAR, DFNT_FLOAT64, DFACC_CREATE, DFACC_WRITE
      implicit none
      real :: error
      integer :: m,n,i
      integer           :: sfsnatt, sfscatt, io, sfstart
      integer           :: sfcreate, sfscompress, sds_id, sfwdata
      integer           :: sfendacc, sfend, dimid3, dimid0, dimid1
      integer           :: dimid2, sfdimid, sfsdmname
      integer           :: region, ind, imr,jmr, lmr
      !    Perhaps implement the function: f(x)=1.0+(x-1.0)*exp(-1./tau) with tau the
      !    e-folding time wanted for each region

      do n=1,nparams_in_state
         state_X(nlag)%xpc(n)=(state_X(nlag-2)%xpc(n)+state_X(nlag-1)%xpc(n)+1.0)/3.0  ! all PE's propagate mean with M=I
      enddo
      !
      !WP! Make the new covariance matrix a combination of the prior and the previous
      !WP! covariance matrix by equal sampling from pdX from previous weeks and current week
      !
      !do m=1,nmembersloc
      ! i=((m-1)*2)/nmembersloc !WP! yields 0,1, or 2 on interval 0:nmembers-1
      ! state_X(nlag)%pdX(:,m)=state_X(nlag-i)%pdX(:,m)
      !enddo

    end subroutine DynamicalModel
    
  end subroutine CreateEnsemble


  subroutine Vectorize(obsonly)

    use user_output_forecast, only: evaluate_forecast

    ! subroutine fills the vector X and matrix Q with the appropriate values from the state
    ! just before inversion
    implicit none
    !i/o
    logical,intent(in),optional :: obsonly
    character(len=*), parameter ::  rname = mname//'/Vectorize'
    integer :: n,m

    ! allocate all variables needed in invert

    if(myid==root) then

       if(.not.present(obsonly)) then 
          allocate(K(nlag*nparams_in_state))
          allocate(corr(nlag*nparams_in_state))
          allocate(meanstate(nlag*nparams_in_state))
          allocate(alpha(n_fcast))
          allocate(dX(nlag*nparams_in_state,nmembers)) !normalized deviations
       endif
       allocate(dF(n_fcast,nmembers)) !normalized deviations

    else

       if(.not.present(obsonly)) then 
          allocate(K(1))
          allocate(corr(1))
          allocate(meanstate(1))
          allocate(alpha(1))
          allocate(dX(1,1))
       endif
       allocate(dF(1,1))

    endif

    call evaluate_forecast

    ! gather each PE's concentration forecast and state deviation to root

#ifdef MPI

    call GatherObs

    if(.not.present(obsonly)) call GatherState

#endif

    ! dX now contains N state vector deviations, non-normalized
    ! dF contains N observations and needs the mean subtracted

    if(myid==root) then

       if(.not.present(obsonly)) then 
          do n=1,nlag
             meanstate( (n-1)*nparams_in_state+1:n*nparams_in_state)=state_X(n)%xpc
          enddo

!          print *,'[DEBUG] Vectorize: meanstate = ',meanstate(1)
       endif
!       print *,'DEBUG: dF(1,1)',dF(1,1)
         do n=nmembers,1,-1
            dF(:,n)=(dF(:,n)-dF(:,1))  ! do n=1 last so mean doesn't change
         enddo

      endif

#ifdef MPI

    contains 

      subroutine GatherObs
        use mpi_const, only: com_trac, my_real, ntracet_ar
        implicit none

        ! local
        integer :: slab1d
        integer,dimension(0:npes-1) :: displ
        integer,dimension(0:npes-1) :: recvcounts
        integer :: sendcount, nsend
        integer :: itracer, i , ierr
        real,allocatable,dimension(:,:) :: sendbuf

        call Par_Barrier
        slab1d    = (n_fcast)
        displ(:)=0  ! offset

        do i=1,npes-1
           displ(i) = displ(i-1) + ntracet_ar(i-1)*slab1d ! calculate offset slabs
        end do
      !
      do i=0,npes-1
         recvcounts(i) = ntracet_ar(i)*slab1d ! receive nr of slabs from other PE's
      end do
      !
      sendcount = nmembersloc*slab1d  ! send the local nr of slabs
      !
      allocate(sendbuf(n_fcast,nmembersloc)) 

      do n=1,nmembersloc
         sendbuf(:,n)=concentration(:)%model(n)
      enddo

      !print*,'Sending ',myid,sum(sendbuf)

      call Par_Barrier
      ! gather on PE root
      call MPI_GATHERV(sendbuf(1,1),sendcount,&
           my_real,dF(1,1),recvcounts,displ,my_real,root,com_trac, ierr)
      if ( ierr /= 0 ) call escape_tm('GatherObs: ERROR')
      call Par_Barrier
      if(myid==root) then
         do n=1,nmembers
            print*,'received ',n,sum(dF(:,n)),ubound(df)
         enddo
         print*,'one row ',dF(:,2)
      endif
     
      deallocate(sendbuf)


    end subroutine GatherObs

    subroutine GatherState
      use mpi_const, only: com_trac, my_real,  ntracet_ar
      implicit none

      ! local
      integer :: slab2d
      integer,dimension(0:npes-1) :: displ
      integer,dimension(0:npes-1) :: recvcounts
      integer :: sendcount, nsend
      integer :: itracer, i, m , ierr
      real,allocatable,dimension(:,:) :: sendbuf

      call Par_Barrier
      slab2d    = (nlag*nparams_in_state)
      displ(:)=0  ! offset

      do i=1,npes-1
         displ(i) = displ(i-1) + ntracet_ar(i-1)*slab2d ! calculate offset slabs
      end do
      !
      do i=0,npes-1
         recvcounts(i) = ntracet_ar(i)*slab2d ! receive nr of slabs from other PE's
      end do
      !
      sendcount = nmembersloc*slab2d  ! send the local nr of slabs
      !
      allocate(sendbuf(nlag*nparams_in_state,nmembersloc)) 

      do n=1,nmembersloc
         do m=1,nlag
            sendbuf((m-1)*nparams_in_state+1:m*nparams_in_state,n)=state_X(m)%pdX(:,n)
         enddo
      enddo

      !print*,'Sending ',myid,sum(sendbuf)

      call Par_Barrier
      ! gather on PE root
      call MPI_GATHERV(sendbuf(1,1),sendcount,&
           my_real,dX(1,1),recvcounts,displ,my_real,root,com_trac, ierr)
      if ( ierr /= 0 ) call escape_tm('GatherState: ERROR')
      call Par_Barrier
      !
      !if(myid==root) then
      ! do n=1,nmembers
      !    print*,'received ',n,sum(dX(:,n)),ubound(dX)
      ! enddo
      ! print*,'one row ',dX(:,2)
      !endif
      !
      deallocate(sendbuf)

    end subroutine GatherState

#endif

  end subroutine Vectorize

  subroutine unVectorize(obsonly)
    ! subroutine fills the state vector and deviations with the appropriate values from the state
    ! just after inversion
#ifdef MPI
    use mpi_const, only: com_trac, my_real, ierr, root
#endif
    implicit none
    !i/o
    logical,intent(in),optional :: obsonly
    integer :: n,m

    if(.not.present(obsonly)) then
       ! scatter state deviation from root

#ifdef MPI

       call ScatterState

       if(myid==root) then
#endif

          do n=1,nlag
             state_X(n)%xpc=meanstate( (n-1)*nparams_in_state+1:n*nparams_in_state)
          enddo

#ifdef MPI
       endif

       do n=1,nlag
          call mpi_bcast(state_X(n)%xpc(1),nparams_in_state,my_real,root,com_trac,ierr)
       enddo

#endif


       ! deallocate all variables used in invert

       deallocate(K)
       deallocate(corr)
       deallocate(meanstate)
       deallocate(alpha)
       deallocate(dX)
    endif
    deallocate(dF)

#ifdef MPI

  contains 

    subroutine ScatterState
      use mpi_const, only: com_trac, my_real,  ntracet_ar
      implicit none

      ! local
      integer :: slab2d
      integer,dimension(0:npes-1) :: displ
      integer,dimension(0:npes-1) :: sendcounts
      integer :: recvcount, nsend
      integer :: itracer, i, m , ierr
      real,allocatable,dimension(:,:) :: recvbuf
      integer :: offsetn=0

      call Par_Barrier
      slab2d    = (nlag*nparams_in_state)
      displ(:)=0  ! offset

      do i=1,npes-1
         displ(i) = displ(i-1) + ntracet_ar(i-1)*slab2d ! calculate offset slabs
      end do
      !
      do i=0,npes-1
         sendcounts(i) = ntracet_ar(i)*slab2d ! receive nr of slabs from other PE's
      end do
      !
      recvcount = nmembersloc*slab2d  ! send the local nr of slabs
      !
      allocate(recvbuf(nlag*nparams_in_state,nmembersloc)) 
      !
      !if(myid==root) then
      ! do n=1,nmembers
      !    print*,'Sending ',n-1,sum(dX(:,n)),ubound(dX)
      ! enddo
      !endif
      !

      call Par_Barrier
      ! gather on PE root
      call MPI_SCATTERV(dX(1,1),sendcounts,displ,&
           my_real,recvbuf(1,1),recvcount,my_real,root,com_trac, ierr)
      if ( ierr /= 0 ) call escape_tm('ScatterState: ERROR')
      call Par_Barrier

      do n=1,nmembersloc
         do m=1,nlag
            state_X(m)%pdX(:,n)=recvbuf((m-1)*nparams_in_state+1:m*nparams_in_state,n)
         enddo
      enddo

      ! if(myid>0) offsetn=sum(ntracet_ar(0:myid-1)) 
      ! do n=1,nmembersloc
      !   print*,'Received ',offsetn+n,sum(recvbuf(:,n))
      ! enddo

      deallocate(recvbuf)

    end subroutine ScatterState

#endif

  end subroutine unVectorize

  subroutine SampleState
    !
    ! SampleState puts the correct flux into co2_flux to be added in a TM5 run,
    ! and loops over nlag weeks of TM5 to produce CO2 concentrations from the
    ! state fluxes. Sampling is done by subroutine get_forecast, where the
    ! sampling protocol (gridbox, time window, interpolation) is also controlled
    !
    use datetime,         only: tau2date 
    use emission_data,    only: increase_itau
    use Dims,             only: itau,itaui,itaue,idatei,idatee,idate,im,jm,lm, nread
    use user_output,      only: station_data, column_data,eddyflux_data, column3d_data
    use global_data,      only: mass_dat
    use go              , only: TDate, NewDate, IncrDate, operator(+), operator(-), rTotal
#ifdef MPI
    use mpi_const,        only: root, my_real, mpi_comm_world, ierr
#endif

    implicit none

    !local
    integer :: n,nbuf,i

    ! setup time interval for this week and transfer to Tdates needed in tracer
    tread1 = NewDate(time6=idate)
    tread2 = NewDate(time6=(/idate(1:3),0,0,0/)) + IncrDate(sec=floor(idate(4)*3600.0/nread)*nread+nread)
    tdyn = NewDate( time6=idate  )

    do n=1,nlag
       ! setup from start time to end of interval  [k*nread,(k+1)*nread]

       itaue=itau ! advance end date
       call increase_itau(itaue,timestep*86400)  ! end of filter
       call tau2date(itaue,idatee)  ! advance end date
       tend = NewDate( time6=idatee )  ! advance end date

       call init_scalefac(n)  ! put scaling factors onto a geographical map

       call checkseed ! should be synched

       station_data=.false.  ! do not produce station output during this run
       column_data=.false.  ! no column output, could be switched on
       column3d_data=.false.  ! no column output, could be switched on
       eddyflux_data=.false.  ! no eddyfluxcolumn output, could be switched on

       call Par_Barrier

       write (gol,'("start new week ...")'); call goPr

       call Tracer(tread1,tread2,tdyn,tend) 

       call Par_Barrier

    enddo  ! TM5 time itau has now been advanced, caution

    call checkseed ! should be synched

    !WP! Save the seed of the random number generator for use in the next
    !WP! cycle. 

    call random_seed(get=sharedseed(1:2))   ! get sharedseed in seed on each PE

  end subroutine SampleState

  subroutine AdvanceBackground
    !
    ! This subroutine takes a final flux estimate and incorporates that into the
    ! background CO2 field by running TM5 over one time step.
    !
    use initexit,                only: Start
    use user_output,             only: station_data, column_data,eddyflux_data, column3d_data
    use datetime,                only: tau2date 
    use emission_data,           only: increase_itau
    use emission_common,         only: reset_emission_averagers
    use Dims,                    only: itau,itaui,itaue,idatei,idatee,idate,im,jm,lm, nread
    use go                     , only: TDate, NewDate, IncrDate, operator(+), operator(-), rTotal
    use user_output_forecast,    only: read_co2_obs,ObsListFFilename
    use user_output_forecast,    only: write_co2_pseudo
    implicit none
    integer      :: n, status

    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/AdvanceBackground'

    write (gol,'("start ...")'); call goPr
    call Start( .true., status ) !WP! true denotes restart not fresh start
    if (status/=0) call escape_tm('ERROR in AdvanceBackground')  

    call Par_Barrier

    ObsListFFilename='obs_final.list'

    call read_co2_obs(status)   ! get co2 observation time and locations for final week
    IF_ERROR_RETURN(status=1)

    itaue=itau
    call increase_itau(itaue,timestep*86400)  ! end of filter
    call tau2date(itaue,idatee)               ! end of filter

    tread1 = NewDate(time6=idate)
    tread2 = NewDate(time6=(/idate(1:3),0,0,0/)) + IncrDate(sec=floor(idate(4)*3600.0/nread)*nread+nread)
    tdyn = NewDate( time6=idate  )
    tend = NewDate( time6=idatee  )

    call init_scalefac(1)  ! put current final state vector into Rfac/GPPfac

    call reset_emission_averagers  ! reset emission averagers to zero

    !WP! Reseed the RNG with the value it had before randomizing
    !WP! during SampleState in order to
    !WP! reconstruct the exact same perturbations we had when forecasting
    !WP! DO NOT ALTER FILE RANDOMSEED AFTER THIS ROUTINE !WP!

    call random_seed(put=savedseed(1:2))   ! put original seed in seed on each PE

    call checkseed ! should be synched

    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'output.station', station_data ,status ,default=.false.)
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'output.column', column_data ,status ,default=.false.)
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'output.column3d', column3d_data ,status ,default=.false.)
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'output.eddyflux', eddyflux_data ,status ,default=.false.)
    IF_NOTOK_RETURN(status=1)
    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)

    call Par_Barrier

    call Tracer(tread1,tread2,tdyn,tend) 

    call Par_Barrier

    if(pseudo) call write_co2_pseudo  

    call checkseed ! should be synched

  end subroutine AdvanceBackground

  subroutine Invert
    !
    ! The analysis equations are solved with the EnSRF algorithm.
    ! 
    implicit none

    !local
    integer :: stat

    if(myid.ne.root) return ! only root does inverse calculations

    if(kfdebug) call save_invert('invert.hdf','diagnostics for inversion',0)

!    print *,'[DEBUG] Invert: meanstate before solve = ',meanstate

    do stat=1,n_fcast

       ! This is handled in read_co2_obs now.
       !call get_R   ! get model-data-mismatch value

       call make_K

       if(concentration(stat)%flag.eq.1.0) cycle  ! do not use this obs if we have set a flag

       if(concentration(stat)%may_localize) call localize

       call solve_enkf

    enddo
!    print *,'[DEBUG] Invert: meanstate after solve = ',meanstate

    if(kfdebug) call save_invert('invert.hdf','diagnostics for inversion',-99)

  contains

    subroutine make_K
      !
      ! calculate the Kalman gain matrix (vector here) for each observation from
      ! the ensemble of CO2 concentrations (dF) and the ensemble of flux deviations
      ! (dX)
      !
      implicit none
      integer              :: i,n
      real                 :: hqhr, res, chi
      real,dimension(:),pointer ::  dF_row

      dF_row => dF(stat,:)

      do i=1,nlag*nparams_in_state
         K(i)=1.0/(nmembers-1)*sum(dX(i,:)*dF_row)     ! this is not yet really K but QHT which has same dims
         corr(i)=correlate(dX(i,:),dF_row,nmembers)
      enddo

      if(kfdebug)print*,'QH created',sum(K)

      hqhr=1.0/(nmembers-1)*sum(dF_row*dF_row)+concentration(stat)%error**2

      concentration(stat)%hqhr=hqhr  ! save for diagnostics

      ! the following lines set the QC flag to 1 if the model-obs is more
      ! than 3x the standard deviation = sqrt(HQH^T+R). Useful for sites
      ! such a HUN which sometimes show enormous spikes that are real,
      ! but cannot be modeled. Inspect the flags after an inversion to
      ! see which data was rejected because the model could not handle
      ! it. Note that MBL sites are always assumed good and can not be
      ! flagged.

      res=concentration(stat)%obs-concentration(stat)%model(1)
      chi=res/sqrt(hqhr)
      if(abs(chi).gt.obs_rejection_threshold.and.concentration(stat)%may_reject) then
         ! concentration(stat)%error=1000.
         concentration(stat)%flag=1.0
      endif

      !if(kfdebug) then
      !print*,'---biased run!---'
      print*,'(z-Hx) obs= ',concentration(stat)%ident,concentration(stat)%obs
      print*,'(z-Hx) mod= ',concentration(stat)%ident,concentration(stat)%model(1)
      print*,'(z-Hx) err= ',concentration(stat)%ident,concentration(stat)%error
      print*,'(z-Hx) flg= ',concentration(stat)%ident,concentration(stat)%flag
      print*,'(z-Hx) dF = ',concentration(stat)%ident,sum(dF(stat,:))/nmembers
      print*,'-----'
      !endif



      if(kfdebug)print*,'HQHplusR created',hqhr

      alpha(stat)=1.0/(1.0+sqrt(concentration(stat)%error**2/hqhr))

      K=K/hqhr   ! now K has really become K by dividing QHT/QHQT+R

      if(kfdebug)print*,'K created',sum(K)

      nullify(dF_row)

    end subroutine make_K

    subroutine localize
      !   use io_save,       only: readtm3hdf
      implicit none
      integer :: i
      real    :: prob
      real    :: tvalue
      !   real,dimension(nparams_in_state,2) :: landmask
      !   real,dimension(nparams_in_state,nparams_in_state) :: distances
      !   real,dimension(nparams_in_state) :: fac
      !   integer :: mloc,is,js,n
      !   real :: corr_scale

      !   call readtm3hdf('regions.hdf',2,nparams_in_state,nparams_in_state,1,0,distances,'dist')
      !
      !   is=nint(concentration(stat)%lon+181)
      !   js=nint(concentration(stat)%lat+91)
      !   mloc=fluxregions(is,js)
      !   print*,myid,'localize ',is,js,mloc,concentration(stat)%lon,concentration(stat)%lat,stat
      !
      ! WP This is a 'block' function implementation of localization

      !fac=1.0
      !where(distances(mloc,:).gt.corr_scale) fac=0.0

      ! WP This is an exponential function implementation of localization
      !   fac=exp(-1.0*distances(mloc,:)/4000.0)
      !
      !   do n=1,nlag
      !      K((n-1)*nparams_in_state+1:n*nparams_in_state)=fac*K((n-1)*nparams_in_state+1:n*nparams_in_state) 
      !   enddo

      ! WP tvalue is the cutoff for a two-sided probability of 0.01 
      ! WP tvalue is the cutoff for a two-sided probability of 0.05 

      select case(nmembers) 
      case(10)
         tvalue=2.67780
         tvalue=2.00856
      case(50)
         tvalue=2.67780
         tvalue=2.00856
      case(75)
         tvalue=2.64298
         tvalue=1.99210
      case(100)
         tvalue=2.62589
         tvalue=1.98397
      case(150)
         tvalue=2.60900
         tvalue=1.97591
      case(200)
         tvalue=2.60064
         tvalue=1.97190
      case(250)
         tvalue=2.59564
         tvalue=1.96950
      case(300)
         tvalue=2.59232
         tvalue=1.96690
      case(350)
         tvalue=2.58995
         tvalue=1.96677
      case(400)
         tvalue=2.58818
         tvalue=1.96591
      case(500)
         tvalue=2.58570
         tvalue=1.96472
      case(550)
         tvalue=2.58480
         tvalue=1.96429
      case default
         call escape_tm('tvalue in localization not known for this ensemble size')

      end select

      do i=1,nlag*nparams_in_state
         prob=corr(i)/sqrt((1.0-corr(i)**2)/(nmembers-2))
         if(abs(prob) < tvalue) K(i)=0.0
      enddo



    end subroutine localize

    real function correlate(x,y,n)
      implicit none
      real,intent(in),dimension(:) :: x,y
      integer, intent(in)          :: n
      real                         :: sum_sq_x,sum_sq_y,sum_coproduct,mean_x,mean_y, delta_x, delta_y, &
           sweep,pop_sd_x,pop_sd_y,cov_x_y
      integer                      :: i

      !correlate=(sum(x*y)-sum(x*y)/n)/sqrt(abs((sum(x**2)-sum(x)**2/n)*(sum(y**2)-sum(y)**2/n)))   
      !correlate=(n*sum(x*y)-sum(x)*sum(y))/sqrt((n*sum(x**2)-sum(x)**2)*(n*sum(y**2)-sum(y)**2))
      sum_sq_x = 0
      sum_sq_y = 0
      sum_coproduct = 0
      mean_x = x(1)
      mean_y = y(1)
      do i=2, N
         sweep = (i - 1.0) / i
         delta_x = x(i) - mean_x
         delta_y = y(i) - mean_y
         sum_sq_x = sum_sq_x + delta_x * delta_x * sweep
         sum_sq_y = sum_sq_y + delta_y * delta_y * sweep
         sum_coproduct = sum_coproduct + delta_x * delta_y * sweep
         mean_x = mean_x + delta_x / i
         mean_y = mean_y + delta_y / i 
      enddo
      pop_sd_x = sqrt( sum_sq_x / N )
      pop_sd_y = sqrt( sum_sq_y / N )
      cov_x_y = sum_coproduct / N
      if(cov_x_y==0.0) then
         correlate=0.0
      else
         correlate = cov_x_y / (pop_sd_x * pop_sd_y)
      endif

    end function correlate


    subroutine solve_enkf
      !
      ! update state using residual and K
      ! update deviations using K and alpha
      !
      use datetime,    only: tstamp
      implicit none
      ! local
      real,dimension(:),  pointer :: dX_col 
      integer :: i, n,m,l, istat
      real    :: res,fac


      ! 1) Invert central value

      res=concentration(stat)%obs-concentration(stat)%model(1)

      meanstate=meanstate+K*res

      if(kfdebug) print*,'new meanstate calculated',sum(meanstate)


      ! 2) Invert ensemble members

      do i=1,nmembers

         dX_col => dX(:,i)

         res=dF(stat,i)

         dX_col=dX_col-alpha(stat)*K*res  ! dX and res are non-normalized quantities, this is okay

         nullify(dX_col)

      enddo

      !WP! Note: this would be the place to check the uncertainty reduction by
      !calculating the stddev for each element of the state before and after the
      !update. We could decide to skip observations that only minimally impact the
      !uncertainty reduction (efficieincy gain for large number of obs), or even decide to do a
      !localization that only allows elements to be updated if the update is
      !significantly large in terms of variance. This is what Dusanka et al do for
      !their filter localization. This addition would require swapping step (1) and
      !(2) which can be done without problems.

      ! 3) Update future central observed value and ensemble observed value

      !do i=stat+1,n_fcast
      do i=1,n_fcast  ! adjust all observations to get an estimate of posterior concentrations

         res=concentration(stat)%obs-concentration(stat)%model(1)  ! the residual vector

         fac=1.0/(nmembers-1)*sum(dF(i,:)*dF(stat,:))/concentration(stat)%hqhr  ! the multiplication factor

         concentration(i)%model(1)=concentration(i)%model(1)+fac*res  ! change in mean conc

         dF(i,:)=dF(i,:)-alpha(stat)*fac*dF(stat,:) ! change in ensemble conc, works on non-normalized quant, okay

      enddo

      if(kfdebug)print*,'Done calculating ensemble states'
      if(kfdebug) call save_invert('invert.hdf','diagnostics for inversion',stat)

    end subroutine solve_enkf

    subroutine save_invert(file_name,msg,flag)
      !
      ! Subroutine saves diagnostics for inversion step
      !
      use io_hdf,      only: io_write
      use io_hdf,      only: DFNT_INT32, DFNT_CHAR, DFNT_FLOAT64, DFACC_CREATE, DFACC_WRITE
      use datetime,    only: tstamp
      implicit none

      ! in/out
      character(len=*),intent(in) :: file_name
      character(len=*),intent(in) :: msg
      integer,intent(in) :: flag

      ! local

      character(len=*), parameter ::  rname = mname//'/Save_Invert'
      integer           :: istat, sfsnatt, sfscatt, io, sfstart
      integer           :: sfcreate, sfscompress, sds_id, sfwdata
      integer           :: sfendacc, sfend, dimid3, dimid0, dimid1
      integer           :: dimid2, sfdimid, sfsdmname
      integer           :: region, ind, imr,jmr, lmr,n,m
      integer           :: status

      character(len=64) :: msg_file

      ind = len_trim(file_name)

      select case(flag)

      case(0)
         io = sfstart(file_name(1:ind),DFACC_CREATE)
         print *, 'save_invert: io unit',io, ' on file ', &
              file_name(1:ind)

         istat = sfscatt(io,'stations', DFNT_CHAR, n_fcast*7, concentration%ident)
         istat = sfscatt(io,'stations_itau', DFNT_INT32, n_fcast, concentration%itau)
         istat = sfscatt(io,'sharedseed', DFNT_INT32, 2 , sharedseed)

         call io_write(io,nlag*nparams_in_state,'m',meanstate,'xp')
         call io_write(io,nlag*nparams_in_state,'m',nmembers,'i',dX,'pX')

         call io_write(io,360,'lon',180,'lat',area_1x1,'area_1x1')
         call io_write(io,nparams_in_state,'nparams_in_state',area_state,'area')
         call io_write(io,n_fcast,'n_fcast',concentration%obs,'obs')   ! obs for timestep itau
         call io_write(io,n_fcast,'n_fcast',concentration%model(1),'model')   ! model for timestep itau
         call io_write(io,n_fcast,'n_fcast',nmembers,'nmembers',dF,'dF') ! ensemble forecasts

      case(-99)

         io = sfstart(file_name(1:ind),DFACC_WRITE)
         call io_write(io,n_fcast,'n',concentration%hqhr,'hqhr') 
         call io_write(io,n_fcast,'n',concentration%flag,'flag')
         call io_write(io,n_fcast,'n',concentration%error,'error')
         call io_write(io,n_fcast,'n',concentration%model(1),'model_post')
         call io_write(io,n_fcast,'n',alpha,'alpha') 
         call io_write(io,nlag*nparams_in_state,'m',meanstate,'x')
         call io_write(io,nlag*nparams_in_state,'m',nmembers,'i',dX,'aX')

      case default 

         io = sfstart(file_name(1:ind),DFACC_WRITE)
         call io_write(io,nlag*nparams_in_state,'m',meanstate, 'xac_'//concentration(flag)%ident)
         call io_write(io,nlag*nparams_in_state,'m',K,'K_'//concentration(flag)%ident)
         call io_write(io,nmembers,'m',df(stat,:),'res_'//concentration(flag)%ident)

      end select

      status=sfend(io)
      write (gol,*)'Save_Invert: sfend returns ',status; call goPr
      if(status/=0) call escape_tm('Stop in '//rname)

    end subroutine save_invert

  end subroutine Invert


  subroutine checkseed
#ifdef MPI
    use mpi_const,              only: myid, mpi_integer,ierr, mpi_comm_world, mpi_max, npes
    integer,dimension(2) :: testseed,totseed
    integer              :: status

    call goLabel(mname//'/checkseed')

    call random_seed(get=testseed(1:2))  

    call mpi_allreduce(testseed,totseed, 2,mpi_integer,mpi_max,mpi_comm_world,ierr)

    if(totseed(1).ne.testseed(1).or.totseed(2).ne.testseed(2)) then
       write (gol,'("Random numbers not synchronized ...")'); call goErr
       write (gol,'("in program")'); call goErr; status=1
       call escape_tm('Stop in checkseed')
    endif
#endif


  end subroutine checkseed


  subroutine SaveEnsemble(df_name,ap_letter,obsdim_name,status)
    !
    ! Subroutine saves ensemble flux perturbations to the savestate file.
    !
    ! df_name:   'dF'
    !            'dF_f'
    !
    ! ap_letter: 'a' anterior
    !            'p' posterior
    !
    use io_hdf
    use enkf_common,     only: savestate_hdf
    use GO_string

    implicit none

    ! in/out
    character(len=*),intent(in) :: df_name
    character(len=*),intent(in) :: obsdim_name
    character(len=1),intent(in) :: ap_letter
    integer,intent(out)         :: status

    character(len=*), parameter :: rname = mname//'/SaveEnsemble'
    integer                     :: n

    real,dimension(:),  pointer :: P
    real,dimension(:,:),  pointer :: A

    character(len=4) :: dsname1,dsname2
    character(len=6) :: dsname

    status = 0

    if(myid.ne.root) return  ! only root writes

    if(.not. allocated(dF)) then
       write (gol,*) 'Inappropriate call the SaveEnsemble; inverse ensemble dF not allocated';call goPr
       status=1
       return
    endif

    call io_write(savestate_hdf%id,n_fcast,obsdim_name,nmembers,'nmembers',dF,df_name,units='ppm') ! ensemble forecasts

    dsname1='x'//ap_letter//'c_'
    dsname2=ap_letter//'dX_'
!    print *,'[DEBUG] saveensemble with ap_letter='//ap_letter//'; meanstate = ', meanstate

    do n=1,nlag   

       P => meanstate( (n-1)*nparams_in_state+1:n*nparams_in_state)
       A => dX(        (n-1)*nparams_in_state+1:n*nparams_in_state,:) 

       dsname=trim(dsname1//goNum2Str(n,'(i2.2)'))
       call io_write(savestate_hdf%id,nparams_in_state,'nparams_in_state',P,dsname,units='dimensionless')

       dsname=trim(dsname2//goNum2Str(n,'(i2.2)'))
       call io_write(savestate_hdf%id,nparams_in_state,'nparams_in_state',nmembers,'nmembers',A,dsname,units='dimensionless')

       nullify(P)
       nullify(A)

    end do  !nlag

  end subroutine SaveEnsemble

  subroutine SaveSeed(status)

    use io_hdf
    use enkf_common,     only: savestate_hdf
    use file_hdf,        only: WriteAttribute

    implicit none

    ! in/out
    integer,intent(out)           :: status

    character(len=*), parameter ::  rname = mname//'/SaveSeed'

    status = 0

    if(myid.ne.root) return  ! only root writes

    call WriteAttribute( savestate_hdf, 'random_seed'      , sharedseed  , status )
    IF_ERROR_RETURN(status=1)

  end subroutine SaveSeed

end module enkf_tools

