!#################################################################
!
! Open and read meteo files.
! Perform some meteo_dependend calculations
!
!### macro's #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################

module meteo

  use GO         , only : gol, goErr, goPr, goLabel
  use GO         , only : TDate
  use grid       , only : TllGridInfo, TLevelInfo
  use tmm        , only : TtmMeteo
  
  use dims       , only : nregions
  
  use meteodata  , only : Set, CoarseMeteo_Init, CoarseMeteo_Done
  use meteodata  , only : TimeInterpolation

  use meteodata  , only : lli, lli_1x1
  use meteodata  , only : levi, levi_ec, levi_tm60

  use meteodata  , only : sp1_dat, sp2_dat, sp_dat, spm_dat
  use meteodata  , only : mfu_dat, mfv_dat, mfw_dat
  use meteodata  , only :  pu_dat,  pv_dat
  use meteodata  , only : temper_dat, humid_dat, gph_dat, omega_dat
  use meteodata  , only : lwc_dat, iwc_dat, cc_dat, cco_dat, ccu_dat
  use meteodata  , only : entu_dat, entd_dat, detu_dat, detd_dat
  use meteodata  , only :    oro_sfc,    oro
  use meteodata  , only : lsmask_sfc, lsmask
  use meteodata  , only : albedo_sfc, albedo
  use meteodata  , only : sr_ecm_sfc, sr_ecm
  use meteodata  , only : sr_ols_sfc, sr_ols
  use meteodata  , only :    sp_sfc
  use meteodata  , only :    ci_sfc,      ci
  use meteodata  , only :   sst_sfc
  use meteodata  , only :   u10m_sfc,   u10m
  use meteodata  , only :   v10m_sfc,   v10m
  use meteodata  , only :   sshf_sfc,   sshf
  use meteodata  , only :   slhf_sfc,   slhf
  use meteodata  , only :     ss_sfc,     ss
  use meteodata  , only :    lsp_sfc,    lsp
  use meteodata  , only :     cp_sfc,     cp

  use meteodata  , only :     sd_sfc
  use meteodata  , only :  swvl1_sfc
  use meteodata  , only :    src_sfc
  use meteodata  , only :    d2m_sfc
  use meteodata  , only :    t2m_sfc
  use meteodata  , only :    ssr_sfc

  use meteodata  , only : nveg
  use meteodata  , only :     tv_sfc
  use meteodata  , only :    cvl_sfc
  use meteodata  , only :    cvh_sfc

  implicit none
  
  ! --- in/out -------------------------------------

  private

  public  ::  Meteo_Init, Meteo_Done, Meteo_Alloc
  public  ::  Meteo_Setup_Mass, Meteo_Setup_Other
  public  ::  Set, CoarseMeteo_Init, CoarseMeteo_Done
  public  ::  TimeInterpolation

  public  ::  sp1_dat, sp2_dat, sp_dat, spm_dat
  public  ::  mfu_dat, mfv_dat, mfw_dat
  public  ::   pu_dat,  pv_dat
  public  ::  temper_dat, humid_dat, gph_dat, omega_dat

  public  ::  use_clouds, new_cloud
  public  ::  lwc_dat, iwc_dat, cc_dat, cco_dat, ccu_dat
  public  ::  entu_dat, entd_dat, detu_dat, detd_dat
  public  ::     oro_sfc,    oro
  public  ::  lsmask_sfc, lsmask
  public  ::  albedo_sfc, albedo
  public  ::  sr_ecm_sfc, sr_ecm
  public  ::  sr_ols_sfc, sr_ols
  public  ::      sp_sfc
  public  ::      ci_sfc,     ci
  public  ::     sst_sfc
  public  ::    u10m_sfc,   u10m
  public  ::    v10m_sfc,   v10m
  public  ::    sshf_sfc,   sshf
  public  ::    slhf_sfc,   slhf
  public  ::      ss_sfc,     ss
  public  ::     lsp_sfc,    lsp
  public  ::      cp_sfc,     cp

  public  ::      sd_sfc
  public  ::   swvl1_sfc
  public  ::     src_sfc
  public  ::     d2m_sfc
  public  ::     t2m_sfc
  public  ::     ssr_sfc

  public  :: nveg
  public  ::      tv_sfc
  public  ::     cvl_sfc
  public  ::     cvh_sfc

  public  ::  lli, lli_1x1
  public  ::  levi, levi_ec, levi_tm60


  ! --- const ----------------------------------
  
  ! module name
  character(len=*), parameter  ::  mname = 'Meteo'
  

  ! --- types -----------------------------------
  
  ! storage for single meteo field:
  
  type TMeteoField
    logical              ::  used
    character(len=16)    ::  name
    character(len=16)    ::  unit
    integer              ::  is(2), js(2), ls(2)   ! shapes
    real, pointer        ::  data(:,:,:)
  end type TMeteoField
  

  ! --- interfaces -----------------------------------
  
  interface Setup
    module procedure Setup_2d
    module procedure Setup_3d
  end interface
  
  interface Coarsen
    module procedure Coarsen_emis
  end interface
  
  ! --- var -------------------------------------

  ! flags if new cloud data has been read; filled from cc%changed etc .
  logical                      ::  new_cloud(nregions) = .false.
  logical                      ::  use_clouds

  ! interface to TM meteo data
  type(TtmMeteo), save         ::  tmmd

  ! source key
  character(len=256)           ::  sourcekey(nregions)
  character(len=256)           ::  sourcekey_1x1
  
  ! write meteo ?
  logical                      ::  write_meteo
  character(len=256)           ::  destkey(nregions)
  logical                      ::  write_meteo_1x1
  character(len=256)           ::  destkey_1x1
  
  ! single cell global surface pressure (region 0)
  real                         ::  sp_region0(1,1)
  


contains


  ! ==========================================================
  ! ===
  ! ===  main routines
  ! ===
  ! ==========================================================

  
  subroutine Meteo_Init( status )
  
    use GO         , only : TrcFile, Init, Done, ReadRc
    use Binas      , only : p_global
    use Grid       , only : Init
    use TMM        , only : Init
    use dims       , only : nregions, region_name
    use dims       , only : xbeg, dx, xref, im
    use dims       , only : ybeg, dy, yref, jm
    use dims       , only : nlon360, nlat180
    use dims       , only : echlevs, lme, a_ec, b_ec
    use dims       , only : lm, lmax_conv
    use meteodata  , only : Init
    use global_data, only : rcfile

    ! --- in/out -------------------------------
        
    integer, intent(out)         ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_Init'
    
    ! --- local -----------------------------
    
    integer             ::  region, n
    integer             ::  imr, jmr, lmr
    integer             ::  halo

    type(TrcFile)       ::  rcF
    character(len=10)   ::  tinterp
    character(len=10)   ::  tinterp_curr
    character(len=10)   ::  tinterp_fg
    character(len=256)  ::  sourcekey_curr
    logical             ::  write_meteo_curr
    character(len=256)  ::  destkey_curr

    integer             ::  iveg
    character(len=4)    ::  sveg

    ! --- begin ----------------------------
    
    call goLabel(rname)
    
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! setup horizontal grids
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    ! global parent, single cell (define mid point and spacing)
    call Init( lli(0), 0.0, 360.0, 1, 0.0, 180.0, 1, status, name='globe' )
    IF_NOTOK_RETURN(status=1)
    
    ! define mid of first west/south cells, spacing, and size:
    do n = 1, nregions
      call Init( lli(n), xbeg(n)+dx/xref(n)/2.0, real(dx)/xref(n), im(n), &
                         ybeg(n)+dy/yref(n)/2.0, real(dy)/yref(n), jm(n), status, &
                         name=region_name(n) )
      IF_NOTOK_RETURN(status=1)
    end do
    !
    call Init( lli_1x1, -179.5, 1.0, 360, -89.5, 1.0, 180, status, name='glb1x1' )
    IF_NOTOK_RETURN(status=1)
    
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! level definition
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    ! setup parent level definition:
    call Init( levi_ec, lme, a_ec, b_ec, status )          ! ecmwf levels
    IF_NOTOK_RETURN(status=1)

    ! setup level definition:
    call Init( levi, levi_ec, echlevs, status )   ! tm half level selection
    IF_NOTOK_RETURN(status=1)

    ! all levels ...    
    call Init( levi_tm60, 'tm60', status )          ! TM 60 levels
    

    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! meteo database
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    ! setup interface to TM meteo:
    call Init( tmmd, rcfile, status )
    IF_NOTOK_RETURN(status=1)
    
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! read settings from rcfile
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ! open rcfile:
    call Init( rcF, rcfile, status )
    IF_NOTOK_RETURN(status=1)

    ! read default temporal interpolations:
    call ReadRc( rcF, 'meteo.tinterp.*', tinterp, status )
    IF_NOTOK_RETURN(status=1)
    
    ! meteo archives; default sourcekeys :
    do n = 1, nregions
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(n)//'.*', sourcekey(n), status )
      IF_NOTOK_RETURN(status=1)
    end do
    call ReadRc( rcF, 'tmm.sourcekey.glb1x1.*', sourcekey_1x1, status )
    IF_NOTOK_RETURN(status=1)

    ! write meteo ? default flags:
    call ReadRc( rcF, 'tmm.output.*', write_meteo, status, .false. )
    IF_NOTOK_RETURN(status=1)
    call ReadRc( rcF, 'tmm.output.surf', write_meteo_1x1, status, write_meteo )
    IF_NOTOK_RETURN(status=1)
    
    ! default destination file names:
    do n = 1, nregions
      call ReadRc( rcF, 'tmm.destkey.'//region_name(n)//'.*', destkey(n), status, 'none' )
      IF_NOTOK_RETURN(status=1)
    end do
    call ReadRc( rcF, 'tmm.destkey.glb1x1.*', destkey_1x1, status, 'none' )
    IF_NOTOK_RETURN(status=1)

    ! close rcfile:
    call Done( rcF, status )
    IF_NOTOK_RETURN(status=1)
    
    
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! define meteo data
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    ! global mean surface pressure
    sp_region0 = p_global
    
    ! setup meteo fields: not in use, not allocated:
    do region = 1, nregions
      imr = im(region)
      jmr = jm(region)
      lmr = lm(region)
    
      !
      ! ** surface pressure *************************************
      !
      
      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.sp', tinterp_curr, status, tinterp )
      IF_ERROR_RETURN(status=1)

      ! check time interpolation:
      select case ( tinterp_curr )
        case ( 'interp6', 'interp6_3', 'interp3' )
          ! ok
        case default
          write (gol,'("unsupported time interpolation for surface pressure:")'); call goErr
          write (gol,'("  requested : ",a)') trim(tinterp_curr); call goErr
          write (gol,'("  supported : interp6, interp6_3, interp3")'); call goErr
          call goErr; status=1; return
      end select
    
      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.sp', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.sp', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.sp', &
                               destkey_curr, status, default=destkey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! two extra horizontal cells
      halo = 2
      
      ! start of interval (copied from sp2_dat):
      call Init( sp1_dat(region), 'sp', 'Pa_sp1', 'computed', &
                     (/1,imr/), (/1,jmr/), halo, (/1,1/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)

      ! end of interval; also reads for sp1 and spm :
      call Init( sp2_dat(region), 'sp', 'Pa_sp2', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,1/), &
                     sourcekey_curr, write_meteo_curr, destkey_curr, status )
      IF_NOTOK_RETURN(status=1)

      ! current pressure:
      call Init( sp_dat(region), 'sp', 'Pa', 'computed', &
                     (/1,imr/), (/1,jmr/), halo, (/1,1/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)

      ! surface pressure at mid of dynamic time interval:
      call Init( spm_dat(region), 'sp', 'Pa', 'computed', &
                     (/1,imr/), (/1,jmr/), halo, (/1,1/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)


      !
      ! ** massfluxes *************************************
      !
      
      ! ~~ vertical

      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.mflux_w', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)
      
      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.mflux_w', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.mflux_w', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.mflux_w', &
                               destkey_curr, status, default=destkey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! no extra cells
      halo = 0
      
      ! vertical flux (kg/s)
      call Init( mfw_dat(region), 'mfw', 'kg/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/0,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey_curr, status )
      IF_NOTOK_RETURN(status=1)
     
      ! ~~ horizontal

      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.mflux_uv', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)
      
      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.mflux_uv', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.mflux_uv', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.mflux_uv', &
                               destkey_curr, status, default=destkey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! NOTE: strange old indexing:
      !   pu_tmpp  -->  pu(0:imr,1:jmr  ,1:lmr)  in  pu_t(0:imr+1,0:jmr+1,0:lmr)
      !                                                         ^ ^     ^ ^       too large !
      !   pv_tmpp  -->  pv(1:imr,1:jmr+1,1:lmr)  in  pv_t(0:imr+1,0:jmr+1,0:lmr)
      !                                                   ^     ^ ^       ^       too large !
      ! The extra cells are implemented below as halo cells.

      ! one extra cell
      halo = 1
      
      !! east/west flux (kg/s)
      !call Init( mfu_dat(region), 'mfu', 'kg/s', tinterp_curr, &
      !               (/0,imr/), (/1,jmr/), halo, (/1,lmr/), &
      !               sourcekey_curr, write_meteo, status, default=destkey_curr )
      !IF_NOTOK_RETURN(status=1)

      !! south/north flux (kg/s)
      !call Init( mfv_dat(region), 'mfv', 'kg/s', tinterp_curr, &
      !               (/1,imr/), (/0,jmr/), halo, (/1,lmr/), &
      !               sourcekey_curr, write_meteo, status, default=destkey_curr )
      !IF_NOTOK_RETURN(status=1)


      ! east/west flux (kg/s)
      call Init( mfu_dat(region), 'mfu', 'kg/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/0,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey_curr, status )
      IF_NOTOK_RETURN(status=1)

      ! south/north flux (kg/s)
      call Init( mfv_dat(region), 'mfv', 'kg/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/0,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey_curr, status )
      IF_NOTOK_RETURN(status=1)


      !! east/west flux (kg/s) : BALANCED
      !call Init( pu_dat(region), 'pu', 'kg/s', 'computed', &
      !               (/0,imr/), (/1,jmr/), halo, (/1,lmr/), &
      !               'no-sourcekey', .false., 'no-destkey', status )
      !IF_NOTOK_RETURN(status=1)
      !
      !! south/north flux (kg/s) : BALANCED
      !call Init( pv_dat(region), 'pv', 'kg/s', 'computed', &
      !               (/1,imr/), (/0,jmr/), halo, (/1,lmr/), &
      !               'no-sourcekey', .false., 'no-destkey', status )
      !IF_NOTOK_RETURN(status=1)
      
      halo = 1

      ! east/west flux (kg/s) : BALANCED
      ! NOTE: data is copied from mfu, thus use same tinterp
      !       for correct allocation of data arrays
      call Init( pu_dat(region), 'pu', 'kg/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/0,lmr/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)

      ! south/north flux (kg/s) : BALANCED
      ! NOTE: data is copied from mfv, thus use same tinterp
      !       for correct allocation of data arrays
      call Init( pv_dat(region), 'pv', 'kg/s',tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/0,lmr/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)


      !
      ! ** temperature *************************************
      !
      
      ! no extra cells
      halo = 0
      
      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.temper', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)

      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.temper', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.temper', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.temper', &
                               destkey_curr, status, default=destkey_curr )
      IF_ERROR_RETURN(status=1)
      
      ! temperature (K)   (halo=0)
      call Init( temper_dat(region), 'T', 'K', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)


      !
      ! ** humidity *************************************
      !
      
      ! no extra cells
      halo = 0
      
      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.humid', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)

      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.humid', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.humid', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.humid', &
                               destkey_curr, status, default=destkey_curr )
      IF_ERROR_RETURN(status=1)
      
      ! humidity (kg/kg)   (halo = 0)
      call Init( humid_dat(region), 'Q', 'kg/kg', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)


      !
      ! ** clouds *************************************
      !
      
      ! no extra cells
      halo = 0
      
      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.cloud', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)

      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.cloud', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.cloud', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)
      
      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.cloud', &
                               destkey_curr, status, default=destkey_curr )
      IF_ERROR_RETURN(status=1)
      
      ! lwc    liquid water content (kg/kg) (halo=0)
      call Init( lwc_dat(region), 'CLWC', 'kg/kg', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! iwc    ice water content (kg/kg) (halo=0)
      call Init( iwc_dat(region), 'CIWC', 'kg/kg', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! cc     cloud cover (fraction) (halo=0)
      call Init( cc_dat(region), 'CC', '0-1', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! cco    cloud cover overhead = above bottom of box (fraction) (halo=0)
      call Init( cco_dat(region), 'CCO', '0-1', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ccu    cloud cover underfeet = below top of box (fraction) (halo=0)
      call Init( ccu_dat(region), 'CCU', '0-1', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

   
      !
      ! ** convection *************************************
      !
      
      ! no extra cells
      halo = 0
      
      ! time interpolation ?  
      call ReadRc( rcF, 'meteo.tinterp.convec', tinterp_curr, status, default=tinterp )
      IF_ERROR_RETURN(status=1)

      ! read sourcekey
      call ReadRc( rcF, 'tmm.sourcekey.'//region_name(region)//'.convec', &
                               sourcekey_curr, status, default=sourcekey(region) )
      IF_ERROR_RETURN(status=1)
      
      ! write this field ?
      call ReadRc( rcF, 'tmm.output.convec', write_meteo_curr, status, default=write_meteo )
      IF_ERROR_RETURN(status=1)

      ! read destkey
      call ReadRc( rcF, 'tmm.destkey.'//region_name(region)//'.convec', &
                               destkey_curr, status, default=destkey_curr )
      IF_ERROR_RETURN(status=1)
      
      ! entu        entrainement updraft
      call Init( entu_dat(region), 'eu', 'kg/m2/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmax_conv/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! entd        entrainement downdraft (im,jm,lmax_conv)
      call Init( entd_dat(region), 'ed', 'kg/m2/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmax_conv/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! detu        detrainement updraft
      call Init( detu_dat(region), 'du', 'kg/m2/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmax_conv/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)

      ! detd        detrainement downdraft
      call Init( detd_dat(region), 'dd', 'kg/m2/s', tinterp_curr, &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmax_conv/), &
                     sourcekey_curr, write_meteo_curr, destkey(region), status )
      IF_NOTOK_RETURN(status=1)


      !
      ! ** computed *************************************
      !
      
      ! no extra cells
      halo = 0
      
      ! geopotential height(m)  (lm+1, halo=0)
      call Init( gph_dat(region), 'gph', 'm', 'computed', &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr+1/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)
      
      ! vertical velocity (Pa/s)  (lm+1, halo=0)
      call Init( omega_dat(region), 'omega', 'Pa/s', 'computed', &
                     (/1,imr/), (/1,jmr/), halo, (/1,lmr+1/), &
                     'no-sourcekey', .false., 'no-destkey', status )
      IF_NOTOK_RETURN(status=1)
      
      ! ***
      
    end do  ! regions
    

    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! surface fields
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    ! no extra cells
    halo = 0

    ! ~~~ orography
    
    ! read sourcekey
    call ReadRc( rcF, 'tmm.sourcekey.glb1x1.oro', sourcekey_curr, status, default=sourcekey_1x1 )
    IF_ERROR_RETURN(status=1)
    
    ! default destkey:
    write_meteo_curr = write_meteo_1x1
    destkey_curr     = destkey_1x1
    
    ! orography (m*[g])
    call Init( oro_sfc, 'oro', 'm m/s2', 'const', &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)
    
    ! ~~~ land/sea mask
    
    ! read sourcekey
    call ReadRc( rcF, 'tmm.sourcekey.glb1x1.lsm', sourcekey_curr, status, default=sourcekey_1x1 )
    IF_ERROR_RETURN(status=1)
    
    ! default destkey:
    write_meteo_curr = write_meteo_1x1
    destkey_curr     = destkey_1x1
    
    ! land/sea mask (%)
    call Init( lsmask_sfc, 'lsm', '%', 'const', &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)
    
    ! ~~~ instantaneous fields
    
    ! time interpolation ?  
    call ReadRc( rcF, 'meteo.tinterp.surf.inst', tinterp_curr, status, default=tinterp )
    IF_ERROR_RETURN(status=1)

    ! default sourcekey:
    sourcekey_curr = sourcekey_1x1

    ! default destkey:
    write_meteo_curr = write_meteo_1x1
    destkey_curr     = destkey_1x1
    
    ! surface pressure on hires grid:
    call Init( sp_sfc, 'sps', 'Pa', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! sea surface temperatue:
    call Init( sst_sfc, 'sst', 'K', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! 10m u wind (m/s; time aver)
    call Init( u10m_sfc, 'u10m', 'm/s', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! 10m v wind (m/s; time aver)
    call Init( v10m_sfc, 'v10m', 'm/s', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    IF(sourcekey_curr(1:4)/='ncep') THEN

        ! skin reservoir content (m water) ; instant
        call Init( src_sfc, 'src', 'm h2o', tinterp_curr, &
                       (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                       sourcekey_curr, .false., destkey_curr, status )
        IF_NOTOK_RETURN(status=1)

        ! 2 meter dewpoint temperature (K)  ; instant
        call Init( d2m_sfc, 'd2m', 'K', tinterp_curr, &
                       (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                       sourcekey_curr, write_meteo_curr, destkey_curr, status )
        IF_NOTOK_RETURN(status=1)

    ENDIF

    ! 2 meter temperature (K)  ; instant
    call Init( t2m_sfc, 't2m', 'K', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! ~~~ average field (accumulated)
    
    ! time interpolation ?  
    call ReadRc( rcF, 'meteo.tinterp.surf.aver', tinterp_curr, status, default=tinterp )
    IF_ERROR_RETURN(status=1)

    ! default sourcekey:
    sourcekey_curr = sourcekey_1x1

    ! default destkey:
    write_meteo_curr = write_meteo_1x1
    destkey_curr     = destkey_1x1
    
    ! surface sensible heat flux (m/s; time aver)
    call Init( sshf_sfc, 'sshf', 'W/m2', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! surface latent heat flux (m/s; time aver)
    call Init( slhf_sfc, 'slhf', 'W/m2', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! surface stress (m/s; time aver)
    call Init( ss_sfc, 'sstr', 'N/m2', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! convective precipitation (m/s; time aver)
    call Init( cp_sfc, 'cp', 'm/s', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! large scale stratiform precipitation (m/s; time aver)
    call Init( lsp_sfc, 'lsp', 'm/s', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! surface solar radiation ( W/m2 ) ; time aver
    call Init( ssr_sfc, 'ssr', 'W/m2', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! ~~~ daily aver fields in tmpp
    
    ! time interpolation ?  
    call ReadRc( rcF, 'meteo.tinterp.surf.day', tinterp_curr, status, default=tinterp )
    IF_ERROR_RETURN(status=1)

    ! default sourcekey:
    sourcekey_curr = sourcekey_1x1

    ! default destkey:
    write_meteo_curr = write_meteo_1x1
    destkey_curr     = destkey_1x1
    
    ! sea ice:
    call Init( ci_sfc, 'ci', '0-1', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! snow depth (m water eqv); day aver ?
    call Init( sd_sfc, 'sd', 'm water eqv', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! volumetric soil water layer 1 ( m3 water / m3 soil) ; day aver ?
    call Init( swvl1_sfc, 'swvl1', 'm3 h2o/m3 soil', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    IF(sourcekey_curr(1:4)/='ncep') THEN
        ! vegetation type (%) ; day aver
        do iveg = 1, nveg
          write (sveg,'("tv",i2.2)') iveg
          call Init( tv_sfc(iveg), sveg, '%', tinterp_curr, &
                         (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                         sourcekey_curr, write_meteo_curr, destkey_curr, status )
          IF_NOTOK_RETURN(status=1)
        end do

        ! low vegetation cover (0-1) ; day aver
        call Init( cvl_sfc, 'cvl', '0-1', tinterp_curr, &
                       (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                       sourcekey_curr, write_meteo_curr, destkey_curr, status )
        IF_NOTOK_RETURN(status=1)

        ! high vegetation cover (0-1) ; day aver
        call Init( cvh_sfc, 'cvh', '0-1', tinterp_curr, &
                       (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                       sourcekey_curr, write_meteo_curr, destkey_curr, status )
        IF_NOTOK_RETURN(status=1)

    ENDIF    

    ! ~~~ albedo
    
    ! time interpolation (first guess) 
    call ReadRc( rcF, 'meteo.tinterp.surf.day', tinterp_fg, status, default=tinterp )
    IF_ERROR_RETURN(status=1)

    ! time interpolation ?  
    call ReadRc( rcF, 'meteo.tinterp.albedo', tinterp_curr, status, default=tinterp_fg )
    IF_ERROR_RETURN(status=1)

    ! read sourcekey
    call ReadRc( rcF, 'tmm.sourcekey.glb1x1.albedo', sourcekey_curr, status, default=sourcekey_1x1 )
    IF_ERROR_RETURN(status=1)

    ! default write flag:    
    write_meteo_curr = write_meteo_1x1

    ! read destkey
    call ReadRc( rcF, 'tmm.destkey.glb1x1.albedo', destkey_curr, status, default=destkey_1x1 )
    IF_ERROR_RETURN(status=1)

    IF(sourcekey_curr(1:4)/='ncep') THEN
        ! albedo ; daily average
        call Init( albedo_sfc, 'albedo', '0-1', tinterp_curr, &
                       (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                       sourcekey_curr, write_meteo_curr, destkey_curr, status )
        IF_NOTOK_RETURN(status=1)

    ENDIF ! ncep-cdc

    ! ~~~ surface roughness
    
    ! time interpolation (first guess)
    call ReadRc( rcF, 'meteo.tinterp.surf.day', tinterp_fg, status, default=tinterp )
    IF_ERROR_RETURN(status=1)

    ! time interpolation ?  
    call ReadRc( rcF, 'meteo.tinterp.sr', tinterp_curr, status, default=tinterp_fg )
    IF_ERROR_RETURN(status=1)

    ! read sourcekey
    call ReadRc( rcF, 'tmm.sourcekey.glb1x1.sr', sourcekey_curr, status, default=sourcekey_1x1 )
    IF_ERROR_RETURN(status=1)
    
    ! default write flag:    
    write_meteo_curr = write_meteo_1x1

    ! read destkey
    call ReadRc( rcF, 'tmm.destkey.glb1x1.sr', destkey_curr, status, default=destkey_1x1 )
    IF_ERROR_RETURN(status=1)

    ! surface roughness (ecmwf,ncep)
    call Init( sr_ecm_sfc, 'sr', 'm', tinterp_curr, &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! ~~~ surface roughness (olsson)
    
    ! defaults
    sourcekey_curr = sourcekey_1x1
    destkey_curr   = destkey_1x1

    ! surface roughness (olsson) ; monthly
    call Init( sr_ols_sfc, 'srols', 'm', 'month', &
                   (/1,nlon360/), (/1,nlat180/), halo, (/1,1/), &
                   sourcekey_curr, .false., destkey_curr, status )
                   !sourcekey_curr, write_meteo_curr, destkey_curr, status )
    IF_NOTOK_RETURN(status=1)

    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! done
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     
    ! ok
    status = 0
    call goLabel()

  end subroutine Meteo_Init


  ! ***


  subroutine Meteo_Done( status )

    use Grid       , only : Done
    use TMM        , only : Done
    use Dims       , only : nregions
    use meteodata  , only : Done

    ! --- in/out -------------------------------
        
    integer, intent(out)         ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_Done'
    
    ! --- local -----------------------------
    
    integer   ::  n
    integer   ::  iveg

    ! --- begin --------------------------------

    call goLabel(rname)
    
    ! interface to TM meteo:
    call Done( tmmd, status )
    IF_NOTOK_RETURN(status=1)
    
    !
    ! done grid definitions
    !
    
    ! horizontal grids:
    do n = 1, nregions
      call Done( lli(n), status )
      IF_NOTOK_RETURN(status=1)
    end do
    !
    call Done( lli_1x1, status )
    IF_NOTOK_RETURN(status=1)
    
    ! done 60level definition:
    call Done( levi_tm60, status )
    IF_NOTOK_RETURN(status=1)

    ! done parent level definition:
    call Done( levi_ec, status )
    IF_NOTOK_RETURN(status=1)

    ! level definition:
    call Done( levi, status )
    IF_NOTOK_RETURN(status=1)

    !
    ! done meteo data
    !

    ! destroy meteo fields:
    do n = 1, nregions

      ! ***
      
      call Done( sp1_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( sp2_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( sp_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( spm_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      ! ***

      call Done( mfu_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( mfv_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( mfw_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( pu_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( pv_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Done( temper_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( humid_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( gph_dat(n), status )
      IF_NOTOK_RETURN(status=1)
    
      call Done( omega_dat(n), status )
      IF_NOTOK_RETURN(status=1)
    
      ! ***
      
      call Done( lwc_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( iwc_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done(  cc_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( cco_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( ccu_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Done( entu_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( entd_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( detu_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      call Done( detd_dat(n), status )
      IF_NOTOK_RETURN(status=1)

    end do
    
    ! *** done surface fields
    
    call Done( oro_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( lsmask_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( albedo_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sr_ecm_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sr_ols_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( ci_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sst_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( u10m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( v10m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( src_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( d2m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( t2m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sshf_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( slhf_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( ss_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( cp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( lsp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( ssr_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( sd_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( swvl1_sfc, status )
    IF_NOTOK_RETURN(status=1)

    do iveg = 1, nveg
      call Done( tv_sfc(iveg), status )
      IF_NOTOK_RETURN(status=1)
    end do

    call Done( cvl_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Done( cvh_sfc, status )
    IF_NOTOK_RETURN(status=1)

    ! ***
    
    ! ok
    status = 0
    call goLabel()

  end subroutine Meteo_Done
  
  
  ! ***


  subroutine Meteo_Alloc( status )
  
    use dims       , only : nregions
    use meteodata  , only : Alloc

    ! --- in/out -------------------------------
        
    integer, intent(out)         ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_Alloc'
    
    ! --- local -----------------------------
    
    integer   ::  region
    integer   ::  iveg

    ! --- begin --------------------------------
    
    call goLabel(rname)
    
    ! allocate meteo fields if in use:
    do region = 1, nregions

      ! ***
      
      call Alloc( sp1_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( sp2_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( sp_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( spm_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Alloc( mfu_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( mfv_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( mfw_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( pu_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( pv_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Alloc( temper_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( humid_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( gph_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( omega_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Alloc( lwc_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( iwc_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc(  cc_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( cco_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( ccu_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
      call Alloc( entu_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( entd_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( detu_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      call Alloc( detd_dat(region), status )
      IF_NOTOK_RETURN(status=1)

      ! ***
      
    end do
        
    ! ***

    call Alloc( oro_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( lsmask_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( albedo_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sr_ecm_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sr_ols_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( ci_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sst_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( u10m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( v10m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( src_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( d2m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( t2m_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sshf_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( slhf_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( ss_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( cp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( lsp_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( ssr_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( sd_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( swvl1_sfc, status )
    IF_NOTOK_RETURN(status=1)

    do iveg = 1, nveg
      call Alloc( tv_sfc(iveg), status )
      IF_NOTOK_RETURN(status=1)
    end do

    call Alloc( cvl_sfc, status )
    IF_NOTOK_RETURN(status=1)

    call Alloc( cvh_sfc, status )
    IF_NOTOK_RETURN(status=1)

    ! ***

    ! ok
    status = 0
    call goLabel()

  end subroutine Meteo_Alloc
  
  

  ! **************************************************************
  ! ***
  ! *** setup
  ! ***
  ! **************************************************************
  
  
  
  subroutine Meteo_Setup_Mass( tr1, tr2, status, isfirst, check_pressure )

    use go       , only : TDate, rTotal, operator(-), wrtgol
    use grid     , only : Match
    use Grid     , only : MassChange, BalanceMassFluxes, CheckMassBalance
    use dims     , only : nregions, im, jm, lm, parent
    use dims     , only : xcyc
    use meteodata, only : SetData

    ! --- in/out -------------------------------
    
    type(TDate), intent(in)        ::  tr1, tr2
    integer, intent(out)           ::  status
    logical, intent(in), optional  ::  check_pressure
    logical, intent(in), optional  ::  isfirst
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_Setup_Mass'
    
    ! --- local -----------------------------
    
    logical                 ::  do_check_pressure
    logical                 ::  do_isfirst
    
    integer                 ::  n, p
    integer                 ::  idater(6)
    real, allocatable       ::  dm_dt(:,:,:)
    real                    ::  dt_sec
    integer                 ::  l
    
    ! --- begin --------------------------------
    
    call goLabel(rname)
    
    ! check pressure ?
    if ( present(check_pressure) ) then
      do_check_pressure = check_pressure
    else
      do_check_pressure = .false.
    end if
    
    ! initial call ?
    if ( present(isfirst) ) then
      do_isfirst = isfirst
    else
      do_isfirst = .false.
    end if
    
    call wrtgol( 'setup mass meteo from: ', tr1, ' to: ', tr2 ); call goPr

    !
    ! ** surface pressures *****************************
    !

    write (gol,'("sp1")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      
      ! skip ?
      if ( .not. sp1_dat(n)%used ) cycle
      
      ! advance 'next' surface pressure to start of interval;
      call Setup( sp2_dat(n), (/tr1,tr1/), lli(n), 'n', status )
      IF_NOTOK_RETURN(status=1)
      ! copy result:
      call SetData( sp1_dat(n), sp2_dat(n), status )
      IF_NOTOK_RETURN(status=1)

      ! global field (first region) ? 
      ! then match with average global surface pressure to ensure global mass balance;
      ! otherwise, match with parent grid:
      if ( n == 1 ) then
        call Match( 'area-aver', 'n', lli(0), sp_region0, &
                                      lli(n), sp1_dat(n)%data(1:im(n),1:jm(n),1), status )
        IF_NOTOK_RETURN(status=1)
      else
        p = parent(n)
        call Match( 'area-aver', 'n', lli(p), sp1_dat(p)%data(1:im(p),1:jm(p),1), &
                                      lli(n), sp1_dat(n)%data(1:im(n),1:jm(n),1), status )
        IF_NOTOK_RETURN(status=1)
      end if
      
      ! initial call ? then set current surface pressure to just read sp1 ;
      ! otherwise, sp remains filled with the advected pressure:
      if ( do_isfirst ) then
        call SetData( sp_dat(n), sp1_dat(n), status )
        IF_NOTOK_RETURN(status=1)
      end if

      ! fill pressure and mass arrays
      call Meteo_SetupMass( n, status )
      IF_NOTOK_RETURN(status=1)

      ! check 'advected' pressure ?
      if ( do_check_pressure) then
        ! compare 'advected' pressure still in sp with just read pressure sp1 :
        call Meteo_CheckPressure( n, status )
        IF_NOTOK_RETURN(status=1)
      end if

    end do  ! regions
    
    ! **

    write (gol,'("sp2")'); call goPr

    ! loop over regions:    
    do n = 1, nregions

      ! skip ?
      if ( .not. sp2_dat(n)%used ) cycle
      
      ! advance 'next' surface pressure to end of interval:
      call Setup( sp2_dat(n), (/tr2,tr2/), lli(n), 'n', status )
      IF_NOTOK_RETURN(status=1)

      ! global field (first region) ? 
      ! then match with average global surface pressure to ensure global mass balance;
      ! otherwise, match with parent grid:
      if ( n == 1 ) then
        call Match( 'area-aver', 'n', lli(0), sp_region0, &
                                      lli(n), sp2_dat(n)%data(1:im(n),1:jm(n),1), status )
        IF_NOTOK_RETURN(status=1)
      else
        p = parent(n)
        call Match( 'area-aver', 'n', lli(p), sp2_dat(p)%data(1:im(p),1:jm(p),1), &
                                      lli(n), sp2_dat(n)%data(1:im(n),1:jm(n),1), status )
        IF_NOTOK_RETURN(status=1)
      end if
      
    end do  ! regions

    !
    ! ** mass fluxes *************************************
    !

    write (gol,'("mfu,mfv")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      
      ! update horizontal u flux (unbalanced)
      call Setup_MFUV( mfu_dat(n), mfv_dat(n), (/tr1,tr2/), lli(n), levi, status )
      IF_NOTOK_RETURN(status=1)

    end do  ! regions

    ! **
      
    write (gol,'("mfw")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      
      ! update vertical flux
      call Setup( mfw_dat(n), (/tr1,tr2/), lli(n), 'n', levi, 'w', status )
      IF_NOTOK_RETURN(status=1)
      
    end do  ! regions

    ! **
      
    write (gol,'("balance")'); call goPr
    
    ! NOTE: since only the surface pressure gradient is used,
    ! it is not necessary to use the data1 and data2 arrays

    ! loop over regions:    
    do n = 1, nregions
    
      ! skip ?
      if ( .not. pu_dat(n)%used ) cycle
      if ( .not. pv_dat(n)%used ) cycle

      ! length of time step between sp1 and sp2:
      dt_sec = rTotal( sp2_dat(n)%tr(1) - sp1_dat(n)%tr(1), 'sec' )

      ! allocate temporary array:
      allocate( dm_dt(im(n),jm(n),lm(n)) )

      ! mass change (kg) :
      call MassChange( dm_dt, lli(n), levi, &
                 sp1_dat(n)%data(1:im(n),1:jm(n),1), &
                 sp2_dat(n)%data(1:im(n),1:jm(n),1), status )
      IF_NOTOK_RETURN(status=1)

      ! mass change (kg/s) :
      dm_dt = dm_dt / dt_sec   ! kg/s

      ! >>> data1 >>>

      ! initial guess for balanced fluxes are unbalanced fluxes:
      pu_dat(n)%data1   = mfu_dat(n)%data1
      pu_dat(n)%filled1 = mfu_dat(n)%filled1
      pu_dat(n)%tr1     = mfu_dat(n)%tr1
      pv_dat(n)%data1   = mfv_dat(n)%data1
      pv_dat(n)%filled1 = mfv_dat(n)%filled1
      pv_dat(n)%tr1     = mfv_dat(n)%tr1

      ! match with parent grid if necessary; note strange indexing:
      !   pu_dat(n)%data1( 0:im(n), 1:jm(n)  , 1:lm(n) )
      !   pv_dat(n)%data1( 1:im(n), 1:jm(n)+1, 1:lm(n) )
      if ( n > 1 ) then
        p = parent(n)
        !
        do l = 1, lm(n)
          call Match( 'sum', 'u', lli(p), pu_dat(p)%data1(0:im(p),1:jm(p),l), &
                                  lli(n), pu_dat(n)%data1(0:im(n),1:jm(n),l), status )
          IF_NOTOK_RETURN(status=1)
        end do
        !
        do l = 1, lm(n)
          call Match( 'sum', 'v', lli(p), pv_dat(p)%data1(1:im(p),1:jm(p)+1,l), &
                                  lli(n), pv_dat(n)%data1(1:im(n),1:jm(n)+1,l), status )
          IF_NOTOK_RETURN(status=1)
        end do
        !
        do l = 0, lm(n)
          call Match( 'sum', 'n', lli(p), mfw_dat(p)%data1(1:im(p),1:jm(p),l), &
                                  lli(n), mfw_dat(n)%data1(1:im(n),1:jm(n),l), status )
          IF_NOTOK_RETURN(status=1)
        end do
        !
      end if

      ! check initial mass balance:
      ! NOTE: strange old indexing:
      !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      call CheckMassBalance( lli(n), &
                               pu_dat(n)%data1(0:im(n),1:jm(n)  ,1:lm(n)), &
                               pv_dat(n)%data1(1:im(n),1:jm(n)+1,1:lm(n)), &
                               sp1_dat(n)%data(1:im(n),1:jm(n),1), &
                               sp2_dat(n)%data(1:im(n),1:jm(n),1), &
                               dt_sec, 1.0e-4, 1.0e-3, status )
      if (status/=0) then
        write (gol,'("initial mass imbalance too large for region ",i2)') n; call goErr
        call goErr; status=1; return
      end if

      ! balance horizontal fluxes:
      ! NOTE: strange old indexing:
      !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      call BalanceMassFluxes( lli(n), &
                               pu_dat(n)%data1(0:im(n),1:jm(n)  ,1:lm(n)), &
                               pv_dat(n)%data1(1:im(n),1:jm(n)+1,1:lm(n)), &
                               mfw_dat(n)%data1, dm_dt, lli(parent(n)), dt_sec, status )
      IF_NOTOK_RETURN(status=1)

      ! check final mass balance:
      ! NOTE: strange old indexing:
      !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
      call CheckMassBalance( lli(n), &
                               pu_dat(n)%data1(0:im(n),1:jm(n)  ,1:lm(n)), &
                               pv_dat(n)%data1(1:im(n),1:jm(n)+1,1:lm(n)), &
                               sp1_dat(n)%data(1:im(n),1:jm(n),1), &
                               sp2_dat(n)%data(1:im(n),1:jm(n),1), &
                               dt_sec, 1.0e-7, 1.0e-6, status )
      if (status/=0) then
        write (gol,'("final mass imbalance too large for region ",i2)') n; call goErr
        call goErr; status=1; return
      end if

      ! periodic boundary
      if ( xcyc(n) == 1 ) then  
        pu_dat(n)%data1(0      ,:,:) = pu_dat(n)%data1(im(n),:,:)
        pu_dat(n)%data1(im(n)+1,:,:) = pu_dat(n)%data1(1    ,:,:)
      end if

      ! >>> data2 >>>

      if ( any((/mfu_dat%filled2,mfv_dat%filled2,mfw_dat%filled2/)) ) then

        ! check ...
        if ( .not. all((/mfu_dat%filled2,mfv_dat%filled2,mfw_dat%filled2/)) ) then
          write (gol,'("either none or all secondary data should be in use:")'); call goErr
          write (gol,'("  mfu_dat%filled2  : ",l1)') mfu_dat%filled2; call goErr
          write (gol,'("  mfv_dat%filled2  : ",l1)') mfv_dat%filled2; call goErr
          write (gol,'("  mfw_dat%filled2  : ",l1)') mfw_dat%filled2; call goErr
          call goErr; status=1; return
        end if

        ! initial guess for balanced fluxes are unbalanced fluxes:
        pu_dat(n)%data2   = mfu_dat(n)%data2
        pu_dat(n)%filled2 = .true.
        pu_dat(n)%tr2     = mfu_dat(n)%tr2
        pv_dat(n)%data2   = mfv_dat(n)%data2
        pv_dat(n)%filled2 = .true.
        pv_dat(n)%tr2     = mfv_dat(n)%tr2

        ! match with parent grid if necessary; note strange indexing:
        !   pu_dat(n)%data2( 0:im(n), 1:jm(n)  , 1:lm(n) )
        !   pv_dat(n)%data2( 1:im(n), 1:jm(n)+1, 1:lm(n) )
        if ( n > 1 ) then
          p = parent(n)
          !
          do l = 1, lm(n)
            call Match( 'sum', 'u', lli(p), pu_dat(p)%data2(0:im(p),1:jm(p),l), &
                                    lli(n), pu_dat(n)%data2(0:im(n),1:jm(n),l), status )
            IF_NOTOK_RETURN(status=1)
          end do
          !
          do l = 1, lm(n)
            call Match( 'sum', 'v', lli(p), pv_dat(p)%data2(1:im(p),1:jm(p)+1,l), &
                                    lli(n), pv_dat(n)%data2(1:im(n),1:jm(n)+1,l), status )
            IF_NOTOK_RETURN(status=1)
          end do
          !
          do l = 0, lm(n)
            call Match( 'sum', 'n', lli(p), mfw_dat(p)%data2(1:im(p),1:jm(p),l), &
                                    lli(n), mfw_dat(n)%data2(1:im(n),1:jm(n),l), status )
            IF_NOTOK_RETURN(status=1)
          end do
          !
        end if

        ! check initial mass balance:
        ! NOTE: strange old indexing:
        !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        call CheckMassBalance( lli(n), &
                                 pu_dat(n)%data2(0:im(n),1:jm(n)  ,1:lm(n)), &
                                 pv_dat(n)%data2(1:im(n),1:jm(n)+1,1:lm(n)), &
                                 sp1_dat(n)%data(1:im(n),1:jm(n),1), &
                                 sp2_dat(n)%data(1:im(n),1:jm(n),1), &
                                 dt_sec, 1.0e-4, 1.0e-3, status )
        if (status/=0) then
          write (gol,'("initial mass imbalance too large for region ",i2)') n; call goErr
          call goErr; status=1; return
        end if

        ! balance horizontal fluxes:
        ! NOTE: strange old indexing:
        !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        call BalanceMassFluxes( lli(n), &
                                 pu_dat(n)%data2(0:im(n),1:jm(n)  ,1:lm(n)), &
                                 pv_dat(n)%data2(1:im(n),1:jm(n)+1,1:lm(n)), &
                                 mfw_dat(n)%data2, dm_dt, lli(parent(n)), dt_sec, status )
        IF_NOTOK_RETURN(status=1)

        ! check final mass balance:
        ! NOTE: strange old indexing:
        !   pu_tmpp  -->  pu(0:im(n),1:jm(n)  ,1:lm(n))  in  pu_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        !   pv_tmpp  -->  pv(1:im(n),1:jm(n)+1,1:lm(n))  in  pv_t(0:im(n)+1,0:jm(n)+1,0:lm(n))
        call CheckMassBalance( lli(n), &
                                 pu_dat(n)%data2(0:im(n),1:jm(n)  ,1:lm(n)), &
                                 pv_dat(n)%data2(1:im(n),1:jm(n)+1,1:lm(n)), &
                                 sp1_dat(n)%data(1:im(n),1:jm(n),1), &
                                 sp2_dat(n)%data(1:im(n),1:jm(n),1), &
                                 dt_sec, 1.0e-7, 1.0e-6, status )
        if (status/=0) then
          write (gol,'("final mass imbalance too large for region ",i2)') n; call goErr
          call goErr; status=1; return
        end if

        ! periodic boundary
        if ( xcyc(n) == 1 ) then  
          pu_dat(n)%data2(0      ,:,:) = pu_dat(n)%data2(im(n),:,:)
          pu_dat(n)%data2(im(n)+1,:,:) = pu_dat(n)%data2(1    ,:,:)
        end if
        
      end if  ! filled2

      ! >>>

      ! clear
      deallocate( dm_dt )

    end do  ! regions

    
    !
    ! ** done **************************************************
    !
    
    ! ok
    status = 0
    call goLabel()

  end subroutine Meteo_Setup_Mass
  

  ! ***
  
  
  subroutine Meteo_Setup_Other( tr1, tr2, status, isfirst )

    use GO         , only : TDate, NewDate, rTotal, wrtgol
    use GO         , only : operator(-), operator(+), operator(/) 
    use GO         , only : InterpolFractions
    use datetime   , only : tau2date
    use tracer_data, only : Par_Check_Domain
    use dims       , only : nregions, im, jm, lm, parent
    use dims       , only : lmax_conv
    use dims,        only : xcyc
    use Dims,        only : czeta
    use meteodata  , only : SetData
    use global_data, only : use_convec
    use global_data, only : conv_dat, region_dat
    use Phys       , only : ConvCloudDim

    ! --- in/out -------------------------------
    
    type(TDate), intent(in)        ::  tr1, tr2
    integer, intent(out)           ::  status
    logical, intent(in), optional  ::  isfirst
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_Setup_Other'
    
    ! --- local -----------------------------
    
    logical                 ::  do_isfirst
    
    integer                 ::  n, p
    integer                 ::  idater(6)
    real, allocatable       ::  dm_dt(:,:,:)
    real                    ::  dt_sec
    
    integer                 ::  i, j, l
    integer                 ::  lsave
    real                    ::  tote, totd, maxe
    real, pointer           ::  dxyp(:) 

    type(TDate)             ::  tmid
    real                    ::  alfa1, alfa2

    integer                 ::  iveg

    ! --- begin --------------------------------
    
    call goLabel(rname)
    
    ! initial call ?
    if ( present(isfirst) ) then
      do_isfirst = isfirst
    else
      do_isfirst = .false.
    end if
    
    call wrtgol( 'setup other meteo from: ', tr1, ' to: ', tr2 ); call goPr

    !
    ! ** orography *****************************
    !

    write (gol,'("oro1x1")'); call goPr
    
    ! read orography (if necessary):
    call Setup( oro_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! coarsen if necessary:
    call Coarsen( oro_sfc, oro, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    write (gol,'("spm")'); call goPr

    !
    ! ** spm **************************************
    !

    ! loop over regions:    
    do n = 1, nregions
      
      ! skip ?
      if ( .not. spm_dat(n)%used ) cycle
      
      ! mid time:
      tmid = tr1 + ( tr2 - tr1 )/2
      
      ! deterimine weights to sp1 and sp2 :
      call InterpolFractions( tmid, sp1_dat(n)%tr(1), sp2_dat(n)%tr(1), alfa1, alfa2, status )
      IF_NOTOK_RETURN(status=1)
      
      ! interpolate:
      spm_dat(n)%data(1:im(n),1:jm(n),1) = &
          alfa1 * sp1_dat(n)%data(1:im(n),1:jm(n),1) + &
          alfa2 * sp2_dat(n)%data(1:im(n),1:jm(n),1)
          
      ! store time:
      spm_dat(n)%tr = (/tr1,tr2/)
      
    end do  ! regions

    !
    ! ** omega **************************************
    !

    write (gol,'("omega")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      
      ! re-compute omega from vertical mass flux:
      call Compute_Omega( omega_dat(n), lli(n), mfw_dat(n), status )
      IF_NOTOK_RETURN(status=1)
      
    end do  ! regions
    

    !
    ! ** temperature and humid **************************************
    !

    write (gol,'("temper and humid")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
    
      ! ncep meteo requires conversion of virtual temperature using humidity ...
      if ( (temper_dat(n)%sourcekey(1:4) == 'ncep') .or. (humid_dat(n)%sourcekey(1:4) == 'ncep') ) then
      
        ! read temperature and humidity (if necessary):
        call Setup_TQ( temper_dat(n), humid_dat(n), (/tr1,tr2/), lli(n), levi, status )
        IF_NOTOK_RETURN(status=1)
        
      else
      
        ! read temperature (if necessary):
        call Setup( temper_dat(n), (/tr1,tr2/), lli(n), 'n', levi, 'n', status )
        IF_NOTOK_RETURN(status=1)
        
        ! read humidity (if necessary):
        call Setup( humid_dat(n), (/tr1,tr2/), lli(n), 'n', levi, 'n', status )
        IF_NOTOK_RETURN(status=1)
        
      end if
    
    end do  ! regions
    

    !
    ! ** gph **************************************
    !

    write (gol,'("gph")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      
      ! re-compute gph from pressure, temperature, and humidity:
      call compute_gph( n, status )
      IF_NOTOK_RETURN(status=1)
      
    end do  ! regions
    

    !
    ! ** clouds **************************************
    !

    write (gol,'("clouds")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      ! read (if necessary):
      call Setup( lwc_dat(n), (/tr1,tr2/), lli(n), 'n', levi, 'n', status )
      IF_NOTOK_RETURN(status=1)
    end do
    
    ! loop over regions:    
    do n = 1, nregions
      ! read (if necessary):
      call Setup( iwc_dat(n), (/tr1,tr2/), lli(n), 'n', levi, 'n', status )
      IF_NOTOK_RETURN(status=1)
    end do
    
    ! loop over regions:    
    do n = 1, nregions
      ! read (if necessary):
      call Setup_CloudCovers( cc_dat(n), cco_dat(n), ccu_dat(n), (/tr1,tr2/), lli(n), levi, status )
      IF_NOTOK_RETURN(status=1)
    end do
    

    !
    ! ** convection **************************************
    !

    write (gol,'("convection")'); call goPr

    ! loop over regions:    
    do n = 1, nregions
      ! read (if necessary):
      call Setup_Convec( entu_dat(n), entd_dat(n), detu_dat(n), detd_dat(n), &
                            omega_dat(n), (/tr1,tr2/), lli(n), levi, status )
      IF_NOTOK_RETURN(status=1)
    end do

    ! ~~ convective clouds

    ! convective clouds in use ?
    if ( use_convec ) then
    
      ! loop over regions:    
      do n = 1, nregions

        ! update necessary ?
        if ( any((/entu_dat(n)%changed,entd_dat(n)%changed/)) ) then

          ! loop over grid cells
          do j = 1, jm(n)
            do i = 1, im(n)

              ! compute convective cloud dimensions for this column:
              call ConvCloudDim( 'u', size(detu_dat(n)%data,3), &
                       detu_dat(n)%data(i,j,:), entd_dat(n)%data(i,j,:),&
                       conv_dat(n)%cloud_base(i,j), &
                       conv_dat(n)%cloud_top (i,j), & 
                       conv_dat(n)%cloud_lfs (i,j), &
                       status )
              IF_NOTOK_RETURN(status=1)

            end do    ! i
          end do   ! j

        end if  ! changed ?

      end do  ! regions

    end if  ! use_convec
    
    ! ~~ unit conversion
      
    ! loop over regions:    
    do n = 1, nregions
    
      ! skip ?
      if ( .not. entu_dat(n)%used ) cycle
      if ( .not. entd_dat(n)%used ) cycle
      if ( .not. detu_dat(n)%used ) cycle
      if ( .not. detd_dat(n)%used ) cycle

      ! update necessary ?
      if ( any((/entu_dat(n)%changed,entd_dat(n)%changed,&
                 detu_dat(n)%changed,detd_dat(n)%changed/)) ) then

        !cmk calculate the rates in kg/gridbox and scale with czeta

        dxyp => region_dat(n)%dxyp

        do j = 1, jm(n)
          do i = 1, im(n)

            ! kg/m2/s -> kg/gridbox/s * scale_factor
            entu_dat(n)%data(i,j,:) = entu_dat(n)%data(i,j,:)*dxyp(j)*czeta
            detu_dat(n)%data(i,j,:) = detu_dat(n)%data(i,j,:)*dxyp(j)*czeta

            ! ensure netto zero tracer transport by updraught in column
            ! (add difference between total entrement and detrement
            ! to level where entrement reaches maximum):
            tote = sum( entu_dat(n)%data(i,j,:) )
            totd = sum( detu_dat(n)%data(i,j,:) )
            maxe = entu_dat(n)%data(i,j,1)  ! changed: reported by PB feb 2003
            lsave = 1
            do l = 2, lmax_conv
              if ( entu_dat(n)%data(i,j,l) > maxe ) then
                maxe = entu_dat(n)%data(i,j,l)
                lsave = l
              end if
            end do
            entu_dat(n)%data(i,j,lsave) = entu_dat(n)%data(i,j,lsave) - tote + totd

            ! kg/m2/s -> kg/gridbox/s * scale_factor
            entd_dat(n)%data(i,j,:) = entd_dat(n)%data(i,j,:)*dxyp(j)*czeta
            detd_dat(n)%data(i,j,:) = detd_dat(n)%data(i,j,:)*dxyp(j)*czeta

            ! ensure netto zero tracer transport by downdraught in column
            ! (add difference between total entrement and detrement
            ! to level where entrement reaches maximum):
            tote = sum( entd_dat(n)%data(i,j,:) )   ! total entrainement
            totd = sum( detd_dat(n)%data(i,j,:) )   ! total detrainement
            maxe = 0.0
            lsave = lmax_conv
            do l = 1, lmax_conv
              if ( entd_dat(n)%data(i,j,l) > maxe ) then
                maxe = entd_dat(n)%data(i,j,l)
                lsave = l
              end if
            end do
            entd_dat(n)%data(i,j,lsave) = entd_dat(n)%data(i,j,lsave) - tote + totd

          end do
        end do

      end if  ! changed ?
        
    end do  ! regions

    !
    ! ** surface fields *****************************
    !

    write (gol,'("surface fields")'); call goPr
    
    ! * lsmask
    ! 
    call Setup( lsmask_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( lsmask_sfc, lsmask, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * albedo
    ! 
    call Setup( albedo_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( albedo_sfc, albedo, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * sr_ecm
    ! 
    call Setup( sr_ecm_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( sr_ecm_sfc, sr_ecm, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * sr_ols
    ! 
    call Setup( sr_ols_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( sr_ols_sfc, sr_ols, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * surface pressure on hires grid:
    ! 
    call Setup( sp_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * sea ice
    ! 
    call Setup( ci_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( ci_sfc, ci, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * sea surface temperature
    ! 
    call Setup( sst_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)


    ! * u10m
    ! 
    call Setup( u10m_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( u10m_sfc, u10m, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * v10m
    ! 
    call Setup( v10m_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( v10m_sfc, v10m, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * skin reservoir content
    ! 
    call Setup( src_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * 2m dewpoint temperature
    ! 
    call Setup( d2m_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * 2m temperature
    ! 
    call Setup( t2m_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * slhf
    ! 
    call Setup( slhf_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( slhf_sfc, slhf, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * sshf
    ! 
    call Setup( sshf_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( sshf_sfc, sshf, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * surface stress
    ! 
    call Setup( ss_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( ss_sfc, ss, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * convective precipitation
    ! 
    call Setup( cp_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( cp_sfc, cp, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * large scale stratiform precipitation
    ! 
    call Setup( lsp_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)
    !
    call Coarsen( lsp_sfc, lsp, 'area-aver', status )
    IF_NOTOK_RETURN(status=1)

    ! * surface solar radiation
    ! 
    call Setup( ssr_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * snow depth
    ! 
    call Setup( sd_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * soil water level 1
    ! 
    call Setup( swvl1_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * vegetation types
    ! 
    do iveg = 1, nveg
      select case ( iveg )
        case ( 1, 2, 3, 4, 5, 6, 7, 9, 10, 11, 13, 16, 17, 18, 19 )
          call Setup( tv_sfc(iveg), (/tr1,tr2/), lli_1x1, 'n', status )
          IF_NOTOK_RETURN(status=1)
        case ( 8, 12, 14, 15, 20 )
          if ( tv_sfc(iveg)%used ) tv_sfc(iveg)%data = 0.0
        case default
          write (gol,'("do not know how to setup vegetation type ",i2)') iveg
          call goErr; status=1; return
      end select
    end do

    ! * low vegetation cover
    ! 
    call Setup( cvl_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)

    ! * high vegetation cover
    ! 
    call Setup( cvh_sfc, (/tr1,tr2/), lli_1x1, 'n', status )
    IF_NOTOK_RETURN(status=1)


    !
    ! ** done ********************************************
    !
    
    ! ok
    status = 0
    call goLabel()

  end subroutine Meteo_Setup_Other
  

  ! **************************************************************
  ! ***
  ! *** setup setup
  ! ***
  ! **************************************************************
  

  subroutine SetupSetup( md, tr, &
                          data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                          data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                          status )
  
    use GO       , only : TDate, NewDate, IncrDate, AnyDate, Get, wrtgol, iTotal
    use GO       , only : operator(+), operator(-), operator(/)
    use GO       , only : operator(==), operator(/=), operator(<), operator(<=)
    use meteodata, only : TMeteoData

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(in)          ::  md
    type(TDate), intent(in)               ::  tr(2)
    logical, intent(out)                  ::  data1_read, data1_copy
    type(TDate), intent(out)              ::  data1_tref, data1_t1, data1_t2
    logical, intent(out)                  ::  data2_read, data2_copy
    type(TDate), intent(out)              ::  data2_tref, data2_t1, data2_t2
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/SetupSetup'
    
    ! --- local ----------------------------------

    integer                         ::  dth, baseh
    integer                         ::  year, month, day, hour, minu
    type(TDate)                     ::  tmid
    type(TDate)                     ::  tc(2)
    integer                         ::  dth_int
    type(TDate)                     ::  tprev, tnext

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! default output:
    data1_read    = .false.
    data1_copy    = .false.
    data2_read    = .false.
    data2_copy    = .false.
    
    !
    ! trap constant fields ...
    !
    
    ! constant and already filled ? then leave
    if ( (md%tinterp == 'const') .and. md%filled1 ) then
      call goLabel(); status = 0; return
    end if
    
    
    !
    ! time stuff
    !
    
    ! basic time resolution in hours
    select case ( md%tinterp )
      case ( 'const', 'month' )
        ! nothing to be set here ...
      case ( 'aver24' )
        ! constant fields produced valid for [00,24]
        dth   = 24
        baseh = 00
      case ( 'aver24_3' )
        ! constant fields produced by tmpp valid for [21,21] = [09-12,09+12]
        dth   = 24
        baseh = -3
      case ( 'const3', 'interp3', 'aver3' )
        dth = 3
        baseh = 0
      case ( 'const6', 'interp6', 'aver6' )
        dth = 6
        baseh = 0
      case ( 'interp6_3' )
        dth = 6
        baseh = 3
      case default
        write (gol,'("unsupported time interpolation : ",a)') md%tinterp; call goErr
        call goErr; status=1; return
    end select

    ! set time parameters for field to be read:
    select case ( md%tinterp )
      !
      ! ** constant fields
      !
      case ( 'const' )
        ! read main field ?
        data1_read = .not. md%filled1
        ! read or leave ?
        if ( data1_read ) then
          data1_tref = tr(1)        !  <--- used for file names
          data1_t1   = AnyDate()
          data1_t2   = AnyDate()
        else
          ! field valid around requested interval, thus leave:
          call goLabel(); status=0; return
        end if
      !
      ! ** constant fields, valid for complete month
      !    
      case ( 'month' )
        ! extract time values for begin of current interval:
        call Get( tr(1), year=year, month=month )
        ! interval for this month:
        tc(1) = NewDate( year=year, month=month, day=01, hour=00 )
        month = month + 1
        if ( month > 12 ) then
          month = 1
          year = year + 1
        end if
        tc(2) = NewDate( year=year, month=month, day=01, hour=00 )
        ! check for strange values:
        if ( (tr(1) < tc(1)) .or. (tc(2) < tr(2)) ) then
          write (gol,'("determined invalid constant interval:")'); call goErr
          call wrtgol( '  requested   : ', tr(1), ' - ', tr(2) ); call goErr
          call wrtgol( '  guessed     : ', tc(1), ' - ', tc(2) ); call goErr
          write (gol,'("  for tinterp : ",a)') md%tinterp; call goErr
          call goErr; status=1; return
          !write (gol,'("          WARNING - requested interval exceeds meteo interval; should be improved")')
        end if
        ! read main field ?
        if ( md%filled1 ) then
          data1_read = md%tr1(1) /= tc(1)
        else
          data1_read = .true.
        end if
        ! read or leave ?
        if ( data1_read ) then
          data1_tref = tr(1)
          data1_t1   = tc(1)
          data1_t2   = tc(2)
        else
          ! field valid around requested interval, thus leave:
          call goLabel(); status=0; return
        end if
      !
      ! ** constant fields, valid for 24hr intervals  [21:00,21:00]
      !    constant fields, valid for  6hr intervals  [21:00,03:00]  etc
      !    constant fields, valid for  3hr intervals  [22:30,01:30]  etc
      !    
      case ( 'const6', 'const3' )
        ! extract time values for begin of current interval:
        call Get( tr(1), year, month, day, hour, minu )
        ! round hour to 00/06/12/18 or 00/03/06/09/12/15/18/21 or 09
        hour = dth * nint(real(hour+minu/60.0-baseh)/real(dth)) + baseh
        ! set mid of 3 or 6 hour interval:
        tmid = NewDate( year, month, day, hour )
        ! interval with constant field
        tc(1) = tmid - IncrDate(hour=dth)/2
        tc(2) = tmid + IncrDate(hour=dth)/2
        ! check for strange values:
        if ( (tr(1) < tc(1)) .or. (tc(2) < tr(2)) ) then
          write (gol,'("determined invalid constant interval:")'); call goErr
          call wrtgol( '  requested   : ', tr(1), ' - ', tr(2) ); call goErr
          call wrtgol( '  guessed     : ', tc(1), ' - ', tc(2) ); call goErr
          write (gol,'("  for tinterp : ",a)') md%tinterp; call goErr
          call goErr; status=1; return
        end if
        ! read main field ?
        if ( md%filled1 ) then
          data1_read = md%tr1(1) /= tmid
        else
          data1_read = .true.
        end if
        ! read or leave ?
        if ( data1_read ) then
          data1_tref = tmid
          data1_t1   = tmid
          data1_t2   = tmid
        else
          ! field valid around requested interval, thus leave:
          call goLabel(); status=0; return
        end if
      !
      ! ** average fields , valid for  3hr intervals  [00:00,03:00]  etc
      !    average fields , valid for  3hr intervals  [00:00,06:00]  etc
      !    
      case ( 'aver3', 'aver6', 'aver24', 'aver24_3' )

        ! extract time values for begin of current interval:
        call Get( tr(1), year, month, day, hour, minu )
        ! round hour to previous  baseh + 00/03/06/09/12/15/18/21 
        hour = dth * floor(real(hour-baseh)/real(dth)) + baseh
        ! interval with constant field
        tc(1) = NewDate( year, month, day, hour )
        tc(2) = tc(1) + IncrDate(hour=dth)
        ! check for strange values:
        if ( (tr(1) < tc(1)) .or. (tc(2) < tr(1)) ) then
          write (gol,'("determined invalid first interval:")'); call goErr
          call wrtgol( '  requested   : ', tr(1), ' - ', tr(2) ); call goErr
          call wrtgol( '  guessed     : ', tc(1), ' - ', tc(2) ); call goErr
          write (gol,'("  for tinterp : ",a)') md%tinterp; call goErr
          call goErr; status=1; return
        end if
        ! read primary field ?
        if ( md%filled1 ) then
          ! read new field if times are different:
          data1_read = (md%tr1(1) /= tc(1)) .or.  (md%tr1(2) /= tc(2))
        else
          ! not filled yet, thus must read:
          data1_read = .true.
        end if
        if ( data1_read ) then
          data1_tref = tc(1)
          data1_t1   = tc(1)
          data1_t2   = tc(2)
        end if

        ! setup reading of secondary data only if end of requested
        !  interval is later than primary interval:
        if ( tc(2) < tr(2) ) then
          ! extract time values for end of requested interval:
          call Get( tr(2), year, month, day, hour, minu )
          ! round hour to next  baseh + 00/03/06/09/12/15/18/21 
          hour = dth * floor(real(hour+minu/60.0-baseh)/real(dth)) + baseh
          ! interval with constant field
          tc(1) = NewDate( year, month, day ) + IncrDate(hour=hour)
          tc(2) = tc(1) + IncrDate(hour=dth)
          ! check for strange values:
          if ( (tr(2) < tc(1)) .or. (tc(2) < tr(2)) ) then
            write (gol,'("determined invalid second interval:")'); call goErr
            call wrtgol( '  requested   : ', tr(1), ' - ', tr(2) ); call goErr
            call wrtgol( '  guessed     : ', tc(1), ' - ', tc(2) ); call goErr
            write (gol,'("  for tinterp : ",a)') md%tinterp; call goErr
            call goErr; status=1; return
          end if
          ! read secondary field ?
          if ( md%filled2 ) then
            ! read new field if times are different;
            data2_read = (md%tr2(1) /= tc(1)) .or. (md%tr2(2) /= tc(2))
          else
            ! not filled yet, thus must read:
            data2_read = .true.
          end if
          if ( data2_read ) then
            data2_tref = tc(1)
            data2_t1   = tc(1)
            data2_t2   = tc(2)
          end if
        end if   ! tr partly after primary interval

      !
      ! ** interpolated between 6 hourly times 00/06/12/18
      !    interpolated between 6 hourly times 03/09/15/21
      !    interpolated between 3 hourly times 00/03/06/09/12/15/18/21
      !
      case ( 'interp6', 'interp6_3', 'interp3' )

        ! extract time values for begin of current interval:
        call Get( tr(1), year, month, day, hour, minu )
        ! truncate hour to previous 00/06/12/18, 03/09/15/21,
        !    or 00/03/06/09/12/15/18/21
        hour = dth * floor(real(hour+minu/60.0-baseh)/real(dth)) + baseh
        ! set begin of 3 or 6 hour interval:
        tprev = NewDate( year, month, day, hour )

        ! extract time values for end of current interval:
        call Get( tr(2), year, month, day, hour, minu )
        ! truncate hour to previous 00/06/12/18
        hour = dth * ceiling(real(hour+minu/60.0-baseh)/real(dth)) + baseh
        ! set end of 3 or 6 hour interval:
        tnext = NewDate( year, month, day, hour )

        ! checks: 
        !   [tprev,tmax] should be dth hours
        !   [tprev,tmax] should contain [tr(1),tr(2)]
        dth_int = iTotal(tnext-tprev,'hour')
        if ( (tr(1) < tprev) .or. (tnext < tr(2)) .or. &
             ( (dth_int /= 0) .and. (dth_int /= dth) ) ) then
          write (gol,'("determined invalid interpolation interval:")'); call goErr
          call wrtgol( '  requested   : ', tr(1), ' - ', tr(2) ); call goErr
          call wrtgol( '  guessed     : ', tprev, ' - ', tnext ); call goErr
          write (gol,'("  for tinterp : ",a)') md%tinterp; call goErr
          call goErr; status=1; return
        end if
        
        !      
        !   .                         <-- previous field at dth hours
        !        o                    <-- latest interpolated field
        !             x               <-- target
        !                  o          <-- next field at dth hours
        !       tr1  tr   tr2
        ! --+--------------+------
        !  tprev          tnext
        !

        ! read main field ?
        if ( md%filled1 ) then
          ! md%data should be defined in [tprev,tr]
          data1_read = (md%tr1(1) < tprev) .or. (tr(2) < md%tr1(1))
        else
          data1_read = .true.
        end if
        if ( data1_read ) then
          data1_tref  = tprev
          data1_t1    = tprev
          data1_t2    = tprev
        end if

        ! read second field ?
        if ( md%filled2 ) then
          ! md%data should be defined for tnext
          data2_read = md%tr2(1) /= tnext
        else
          data2_read = .true.
        end if
        if ( data2_read ) then
          data2_tref = tnext
          data2_t1   = tnext
          data2_t2   = tnext
        end if
        
      !
      ! ** error ...
      !
      case default
        write (gol,'("unsupported time interpolation : ",a)') md%tinterp ; call goErr
        call goErr; status=1; return
    end select
    
    
    !
    ! trap double reading
    !
    
    ! data already in data2 ?
    if ( data1_read .and. md%filled2 ) then
      if ( (data1_t1 == md%tr2(1)) .and. (data1_t2 == md%tr2(2)) ) then
        data1_read = .false.
        data1_copy = .true.
      end if
    end if
    
    ! data2 just read ?
    if ( data2_read .and. data1_read ) then
      ! data2 is same as data ? 
      if ( (data2_tref == data1_tref) .and. &
           (data2_t1 == data1_t1) .and. (data2_t2 == data1_t2) ) then
        data2_read = .false.
        data2_copy = .true.
      end if
    end if
    
    ! ok
    status = 0
    call goLabel()
    
  end subroutine SetupSetup
  
  
   ! **************************************************************
  
  
  subroutine Setup_2d( md, tr, lli, nuv, status )

    use GO       , only : TDate, wrtgol
    use Grid     , only : TllGridInfo
    use TMM      , only : ReadField, Read_SP, Read_SR_OLS, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  md
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    character(len=1), intent(in)          ::  nuv
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_2d'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    !write (gol,'(a)') trim(md%name); call goPr
    
    ! leave if not in use:
    if ( .not. md%used ) then
      call goLabel(); status=0; return
    end if
    
    ! not changed by default
    md%changed = .false.

    !
    ! time stuff
    !

    call SetupSetup( md, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
    
    !
    ! read/write field
    !
    
    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO

        ! fill data:
        select case ( md%name )
          !
          ! special routine for surface pressure
          !
          case ( 'sp' )

            call Read_SP( tmmd, md%sourcekey, &
                            data1_tref, data1_t1, data1_t2, &
                            lli, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                            md%tmi1, status )
            IF_NOTOK_RETURN(status=1)

          !
          ! special routine for Olsson surface roughness:
          !
          case ( 'srols' )

            call Read_SR_OLS( tmmd, md%sourcekey, &
                            data1_tref, data1_t1, data1_t2, &
                            lli, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                            md%tmi1, status )
            IF_NOTOK_RETURN(status=1)

          ! 
          ! general field
          !
          case default

            call ReadField( tmmd, md%sourcekey, trim(md%name), &
                              data1_tref, data1_t1, data1_t2, &
                              lli, nuv, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                              md%tmi1, status )
            IF_NOTOK_RETURN(status=1)

        end select

        ! write meteofiles ?
        if ( md%putout ) then
          call WriteField( tmmd, md%destkey, &
                   md%tmi1, trim(md%name), trim(md%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, nuv, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      
      ! data array is filled now:
      md%filled1  = .true.
      md%tr1(1)   = data1_t1
      md%tr1(2)   = data1_t2
      md%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
      md%data1 = md%data2
      
      ! data array is filled now:
      md%filled1  = .true.
      md%tr1(1)   = data1_t1
      md%tr1(2)   = data1_t2
      md%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO

        ! fill data:
        select case ( md%name )
          !
          ! special routine for surface pressure
          !
          case ( 'sp' )

            call Read_SP( tmmd, md%sourcekey, &
                            data2_tref, data2_t1, data2_t2, &
                            lli, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                            md%tmi2, status )
            IF_NOTOK_RETURN(status=1)

          ! 
          ! general field
          !
          case default

            call ReadField( tmmd, md%sourcekey, trim(md%name), &
                              data2_tref, data2_t1, data2_t2, &
                              lli, nuv, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                              md%tmi2, status )
            IF_NOTOK_RETURN(status=1)

        end select

        ! write meteofiles ?
        if ( md%putout ) then
          call WriteField( tmmd, md%destkey, &
                   md%tmi2, trim(md%name), trim(md%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, nuv, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),1), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      
      ! data array is filled now:
      md%filled2 = .true.
      md%tr2(1)  = data2_t1
      md%tr2(2)  = data2_t2

    else if ( data2_copy ) then
    
      ! copy data from secondary array:
      md%data2 = md%data1
      
      ! data array is filled now:
      md%filled2 = .true.
      md%tr2(1)  = data2_t1
      md%tr2(2)  = data2_t2

    end if
    
    
    !
    ! time interpolation
    !
    
    ! apply time interpolation:
    call TimeInterpolation( md, tr, status )
    IF_NOTOK_RETURN(status=1)

    
    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_2d


  ! ***
  

  subroutine Setup_3d( md, tr, lli, nuv, levi, nw, status )

    use GO       , only : TDate, wrtgol, operator(/=)
    use Grid     , only : TllGridInfo, TLevelInfo
    use TMM      , only : TMeteoInfo, ReadField, Read_MFW, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  md
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    character(len=1), intent(in)          ::  nuv
    type(TLevelInfo), intent(in)          ::  levi
    character(len=1), intent(in)          ::  nw
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_3d'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2
    
    real, allocatable               ::  tmp_sp(:,:)

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! leave if not in use:
    if ( .not. md%used ) then
      call goLabel(); status=0; return
    end if
    
    ! not changed by default
    md%changed = .false.

    !
    ! time stuff
    !

    call SetupSetup( md, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
           
    !
    ! read/write field
    !

    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data1_t2 /= data1_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data1_t1  : ', data1_t1  ); call goErr
          call wrtgol( '  data1_t2  : ', data1_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(md%is(1):md%is(2),md%js(1):md%js(2)) )

        ! fill data:
        select case ( md%name )
          !
          ! special routine for vertical flux:
          !
          case ( 'mfw' )

            call Read_MFW( tmmd, md%sourcekey, &
                                data1_tref, data1_t1, data1_t2, &
                                lli, levi, &
                                tmp_sp, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                                md%tmi1, status )
            IF_NOTOK_RETURN(status=1)

          ! 
          ! general field
          !
          case default

            call ReadField( tmmd, md%sourcekey, trim(md%name), &
                       data1_tref, data1_t1, data1_t2, &
                       lli, nuv, levi, nw, &
                       tmp_sp, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                       md%tmi1, status )
            IF_NOTOK_RETURN(status=1)
            
        end select

        ! write meteofiles ?
        if ( md%putout ) then
          call WriteField( tmmd, md%destkey, &
                   md%tmi1, 'sp', trim(md%name), trim(md%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, nuv, levi, nw, &
                   tmp_sp, md%data1(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md%data1, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md%filled1  = .true.
      md%tr1(1)   = data1_t1
      md%tr1(2)   = data1_t2
      md%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
      md%data1 = md%data2
      
      ! data array is filled now:
      md%filled1  = .true.
      md%tr1(1)   = data1_t1
      md%tr1(2)   = data1_t2
      md%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO

        ! safety check ...
        if ( data2_t2 /= data2_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data2_t1  : ', data2_t1  ); call goErr
          call wrtgol( '  data2_t2  : ', data2_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:      
        allocate( tmp_sp(md%is(1):md%is(2),md%js(1):md%js(2)) )

        ! fill data:
        select case ( md%name )
          !
          ! special routine for vertical flux:
          !
          case ( 'mfw' )

            call Read_MFW( tmmd, md%sourcekey, &
                       data2_tref, data2_t1, data2_t2, &
                       lli, levi, &
                       tmp_sp, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                       md%tmi2, status )
            IF_NOTOK_RETURN(status=1)

          ! 
          ! general field
          !
          case default

            call ReadField( tmmd, md%sourcekey, trim(md%name), &
                       data2_tref, data2_t1, data2_t2, &
                       lli, nuv, levi, nw, &
                       tmp_sp, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                       md%tmi2, status )
            IF_NOTOK_RETURN(status=1)
            
        end select

        ! write meteofiles ?
        if ( md%putout ) then
          call WriteField( tmmd, md%destkey, &
                   md%tmi2, 'sp', trim(md%name), trim(md%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, nuv, levi, nw, &
                   tmp_sp, md%data2(md%is(1):md%is(2),md%js(1):md%js(2),md%ls(1):md%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if

        ! clear
        deallocate( tmp_sp )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md%data2, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md%filled2 = .true.
      md%tr2(1)  = data2_t1
      md%tr2(2)  = data2_t2

    else if ( data2_copy ) then
    
      ! copy data from secondary array:
      md%data2 = md%data1
      
      ! data array is filled now:
      md%filled2 = .true.
      md%tr2(1)  = data2_t1
      md%tr2(2)  = data2_t2

    end if
    
    
    !
    ! time interpolation 
    !
    
    ! apply time interpolation:
    call TimeInterpolation( md, tr, status )
    IF_NOTOK_RETURN(status=1)
      
    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_3d


  ! **************************************************************
  ! ***
  ! *** mass fluxes
  ! ***
  ! **************************************************************
  

  subroutine Setup_MFUV( md_mfu, md_mfv, tr, lli, levi, status )

    use GO       , only : TDate, wrtgol, operator(/=)
    use Grid     , only : TllGridInfo, TLevelInfo
    use TMM      , only : TMeteoInfo, Read_MFUV, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  md_mfu
    type(TMeteoData), intent(inout)       ::  md_mfv
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    type(TLevelInfo), intent(in)          ::  levi
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_MFUV'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2

    real, allocatable               ::  tmp_spu(:,:)
    real, allocatable               ::  tmp_spv(:,:)

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! leave if not in use:
    if ( md_mfu%used .neqv. md_mfv%used ) then
      write (gol,'("either none or both mfu and mfv should be in use")'); call goErr
      call goErr; status=1; return
    end if
    if ( .not. md_mfu%used ) then
      call goLabel(); status=0; return
    end if
    
    ! not changed by default
    md_mfu%changed = .false.
    md_mfv%changed = .false.

    !
    ! time stuff
    !

    ! (sufficient to setup from mfu only)
    call SetupSetup( md_mfu, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
           
    !
    ! read/write field
    !

    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data1_t2 /= data1_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data1_t1  : ', data1_t1  ); call goErr
          call wrtgol( '  data1_t2  : ', data1_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_spu(md_mfu%is(1)-1:md_mfu%is(2),md_mfu%js(1):md_mfu%js(2)  ) )
        allocate( tmp_spv(md_mfv%is(1)  :md_mfv%is(2),md_mfv%js(1):md_mfv%js(2)+1) )

        ! NOTE: strange old indexing:
        !   pu_tmpp  -->  pu(0:imr,1:jmr  ,1:lmr)  in  pu_t(0:imr+1,0:jmr+1,0:lmr)
        !   pv_tmpp  -->  pv(1:imr,1:jmr+1,1:lmr)  in  pv_t(0:imr+1,0:jmr+1,0:lmr)

        ! fill data:
        call Read_MFUV( tmmd, md_mfu%sourcekey, &
                 data1_tref, data1_t1, data1_t2, lli, levi, &
                 tmp_spu, &
                   md_mfu%data1(md_mfu%is(1)-1:md_mfu%is(2),&
                                md_mfu%js(1)  :md_mfu%js(2),&
                                md_mfu%ls(1)+1:md_mfu%ls(2)), &
                   md_mfu%tmi1, &
                 tmp_spv, &
                   md_mfv%data1(md_mfv%is(1)  :md_mfv%is(2)  , &
                                md_mfv%js(1)  :md_mfv%js(2)+1, &
                                md_mfv%ls(1)+1:md_mfv%ls(2)   ), &
                   md_mfv%tmi1, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( md_mfu%putout ) then
          call WriteField( tmmd, md_mfu%destkey, &
                   md_mfu%tmi1, 'spu', trim(md_mfu%name), trim(md_mfu%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'u', levi, 'n', &
                   tmp_spu, md_mfu%data1(md_mfu%is(1)-1:md_mfu%is(2),&
                                         md_mfu%js(1)  :md_mfu%js(2),&
                                         md_mfu%ls(1)+1:md_mfu%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( md_mfv%putout ) then
          call WriteField( tmmd, md_mfv%destkey, &
                   md_mfv%tmi1, 'spv', trim(md_mfv%name), trim(md_mfv%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'v', levi, 'n', &
                   tmp_spv, md_mfv%data1(md_mfv%is(1)  :md_mfv%is(2)  , &
                                         md_mfv%js(1)  :md_mfv%js(2)+1, &
                                         md_mfv%ls(1)+1:md_mfv%ls(2)   ), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_spu )
        deallocate( tmp_spv )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md_mfu%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( md_mfv%data1, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md_mfu%filled1  = .true.
      md_mfu%tr1(1)   = data1_t1
      md_mfu%tr1(2)   = data1_t2
      md_mfu%changed  = .true.
      md_mfv%filled1  = .true.
      md_mfv%tr1(1)   = data1_t1
      md_mfv%tr1(2)   = data1_t2
      md_mfv%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
      md_mfu%data1 = md_mfu%data2
      md_mfv%data1 = md_mfv%data2
      
      ! data array is filled now:
      md_mfu%filled1  = .true.
      md_mfu%tr1(1)   = data1_t1
      md_mfu%tr1(2)   = data1_t2
      md_mfu%changed  = .true.
      md_mfv%filled1  = .true.
      md_mfv%tr1(1)   = data1_t1
      md_mfv%tr1(2)   = data1_t2
      md_mfv%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO

        ! safety check ...
        if ( data2_t2 /= data2_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data2_t1  : ', data2_t1  ); call goErr
          call wrtgol( '  data2_t2  : ', data2_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:      
        allocate( tmp_spu(md_mfu%is(1)-1:md_mfu%is(2),md_mfu%js(1):md_mfu%js(2)  ) )
        allocate( tmp_spv(md_mfv%is(1)  :md_mfv%is(2),md_mfv%js(1):md_mfv%js(2)+1) )

        ! NOTE: strange old indexing:
        !   pu_tmpp  -->  pu(0:imr,1:jmr  ,1:lmr)  in  pu_t(0:imr+1,0:jmr+1,0:lmr)
        !   pv_tmpp  -->  pv(1:imr,1:jmr+1,1:lmr)  in  pv_t(0:imr+1,0:jmr+1,0:lmr)

        ! fill data:
        call Read_MFUV( tmmd, md_mfu%sourcekey, &
                 data2_tref, data2_t1, data2_t2, lli, levi, &
                 tmp_spu, &
                   md_mfu%data2(md_mfu%is(1)-1:md_mfu%is(2),&
                                md_mfu%js(1)  :md_mfu%js(2),&
                                md_mfu%ls(1)+1:md_mfu%ls(2)), &
                   md_mfu%tmi2, &
                 tmp_spv, &
                   md_mfv%data2(md_mfv%is(1)  :md_mfv%is(2)  , &
                                md_mfv%js(1)  :md_mfv%js(2)+1, &
                                md_mfv%ls(1)+1:md_mfv%ls(2)   ), &
                   md_mfv%tmi2, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( md_mfu%putout ) then
          call WriteField( tmmd, md_mfu%destkey, &
                   md_mfu%tmi2, 'spu', trim(md_mfu%name), trim(md_mfu%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'u', levi, 'n', &
                   tmp_spu, md_mfu%data2(md_mfu%is(1)-1:md_mfu%is(2),&
                                         md_mfu%js(1)  :md_mfu%js(2),&
                                         md_mfu%ls(1)+1:md_mfu%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        endif
        if ( md_mfv%putout ) then
          call WriteField( tmmd, md_mfv%destkey, &
                   md_mfv%tmi2, 'spv', trim(md_mfv%name), trim(md_mfv%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'v', levi, 'n', &
                   tmp_spv, md_mfv%data2(md_mfv%is(1)  :md_mfv%is(2)  , &
                                         md_mfv%js(1)  :md_mfv%js(2)+1, &
                                         md_mfv%ls(1)+1:md_mfv%ls(2)   ), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if

        ! clear
        deallocate( tmp_spu )
        deallocate( tmp_spv )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md_mfu%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( md_mfv%data2, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md_mfu%filled2 = .true.
      md_mfu%tr2(1)  = data2_t1
      md_mfu%tr2(2)  = data2_t2
      md_mfv%filled2 = .true.
      md_mfv%tr2(1)  = data2_t1
      md_mfv%tr2(2)  = data2_t2

    else if ( data2_copy ) then
    
      ! copy data from primary array:
      md_mfu%data2 = md_mfu%data
      md_mfv%data2 = md_mfv%data
      
      ! data array is filled now:
      md_mfu%filled2 = .true.
      md_mfu%tr2(1)  = data2_t1
      md_mfu%tr2(2)  = data2_t2
      md_mfv%filled2 = .true.
      md_mfv%tr2(1)  = data2_t1
      md_mfv%tr2(2)  = data2_t2

    end if
    
    
    !
    ! time interpolation
    !
    
    ! apply time interpolation:
    call TimeInterpolation( md_mfu, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( md_mfv, tr, status )
    IF_NOTOK_RETURN(status=1)
    
    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_MFUV



  ! **************************************************************
  ! ***
  ! ***  temperature and humidity
  ! ***
  ! **************************************************************
  

  subroutine Setup_TQ( md_T, md_Q, tr, lli, levi, status )

    use GO       , only : TDate, wrtgol, operator(/=)
    use Grid     , only : TllGridInfo, TLevelInfo
    use TMM      , only : TMeteoInfo, Read_TQ, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  md_T
    type(TMeteoData), intent(inout)       ::  md_Q
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    type(TLevelInfo), intent(in)          ::  levi
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_TQ'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2

    real, allocatable               ::  tmp_sp(:,:)

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! leave if not in use:
    if ( md_T%used .neqv. md_Q%used ) then
      write (gol,'("either none or both T and Q should be in use")'); call goErr
      call goErr; status=1; return
    end if
    if ( .not. md_T%used ) then
      call goLabel(); status=0; return
    end if
    
    ! not changed by default
    md_T%changed = .false.
    md_Q%changed = .false.

    !
    ! time stuff
    !

    ! (sufficient to setup from T only)
    call SetupSetup( md_T, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
           
    !
    ! read/write field
    !

    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data1_t2 /= data1_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data1_t1  : ', data1_t1  ); call goErr
          call wrtgol( '  data1_t2  : ', data1_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(md_T%is(1):md_T%is(2),md_T%js(1):md_T%js(2)  ) )

        ! fill data:
        call Read_TQ( tmmd, md_T%sourcekey, md_Q%sourcekey, &
                 data1_tref, data1_t1, data1_t2, lli, levi, &
                 tmp_sp, &
                   md_T%data1(md_T%is(1):md_T%is(2),&
                              md_T%js(1):md_T%js(2),&
                              md_T%ls(1):md_T%ls(2)), &
                   md_T%tmi1, &
                   md_Q%data1(md_Q%is(1):md_Q%is(2), &
                              md_Q%js(1):md_Q%js(2), &
                              md_Q%ls(1):md_Q%ls(2) ), &
                   md_Q%tmi1, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( md_T%putout ) then
          call WriteField( tmmd, md_T%destkey, &
                   md_T%tmi1, 'sp', trim(md_T%name), trim(md_T%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, md_T%data1(md_T%is(1):md_T%is(2),&
                                     md_T%js(1):md_T%js(2),&
                                     md_T%ls(1):md_T%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( md_Q%putout ) then
          call WriteField( tmmd, md_Q%destkey, &
                   md_Q%tmi1, 'sp', trim(md_Q%name), trim(md_Q%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, md_Q%data1(md_Q%is(1):md_Q%is(2), &
                                      md_Q%js(1):md_Q%js(2), &
                                      md_Q%ls(1):md_Q%ls(2) ), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md_T%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( md_Q%data1, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md_T%filled1  = .true.
      md_T%tr1(1)   = data1_t1
      md_T%tr1(2)   = data1_t2
      md_T%changed  = .true.
      md_Q%filled1  = .true.
      md_Q%tr1(1)   = data1_t1
      md_Q%tr1(2)   = data1_t2
      md_Q%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
      md_T%data1 = md_T%data2
      md_Q%data1 = md_Q%data2
      
      ! data array is filled now:
      md_T%filled1  = .true.
      md_T%tr1(1)   = data1_t1
      md_T%tr1(2)   = data1_t2
      md_T%changed  = .true.
      md_Q%filled1  = .true.
      md_Q%tr1(1)   = data1_t1
      md_Q%tr1(2)   = data1_t2
      md_Q%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO

        ! safety check ...
        if ( data2_t2 /= data2_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data2_t1  : ', data2_t1  ); call goErr
          call wrtgol( '  data2_t2  : ', data2_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:      
        allocate( tmp_sp(md_Q%is(1):md_Q%is(2),md_Q%js(1):md_Q%js(2)) )

        ! fill data:
        call Read_TQ( tmmd, md_T%sourcekey, md_Q%sourcekey, &
                 data2_tref, data2_t1, data2_t2, lli, levi, &
                 tmp_sp, &
                   md_T%data2(md_T%is(1):md_T%is(2),&
                              md_T%js(1):md_T%js(2),&
                              md_T%ls(1):md_T%ls(2)), &
                   md_T%tmi2, &
                   md_Q%data2(md_Q%is(1):md_Q%is(2), &
                              md_Q%js(1):md_Q%js(2), &
                              md_Q%ls(1):md_Q%ls(2) ), &
                   md_Q%tmi2, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( md_T%putout ) then
          call WriteField( tmmd, md_T%destkey, &
                   md_T%tmi2, 'sp', trim(md_T%name), trim(md_T%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, md_T%data2(md_T%is(1):md_T%is(2),&
                                      md_T%js(1):md_T%js(2),&
                                      md_T%ls(1):md_T%ls(2)), &
                   status )
          IF_NOTOK_RETURN(status=1)
        endif
        if ( md_Q%putout ) then
          call WriteField( tmmd, md_Q%destkey, &
                   md_Q%tmi2, 'sp', trim(md_Q%name), trim(md_Q%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, md_Q%data2(md_Q%is(1):md_Q%is(2), &
                                      md_Q%js(1):md_Q%js(2), &
                                      md_Q%ls(1):md_Q%ls(2) ), &
                   status )
          IF_NOTOK_RETURN(status=1)
        end if

        ! clear
        deallocate( tmp_sp )

      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( md_T%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( md_Q%data2, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      md_T%filled2 = .true.
      md_T%tr2(1)  = data2_t1
      md_T%tr2(2)  = data2_t2
      md_Q%filled2 = .true.
      md_Q%tr2(1)  = data2_t1
      md_Q%tr2(2)  = data2_t2

    else if ( data2_copy ) then
    
      ! copy data from primary array:
      md_T%data2 = md_T%data1
      md_Q%data2 = md_Q%data1
      
      ! data array is filled now:
      md_T%filled2 = .true.
      md_T%tr2(1)  = data2_t1
      md_T%tr2(2)  = data2_t2
      md_Q%filled2 = .true.
      md_Q%tr2(1)  = data2_t1
      md_Q%tr2(2)  = data2_t2

    end if
    
    
    !
    ! time interpolation
    !
    
    ! apply time interpolation:
    call TimeInterpolation( md_T, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( md_Q, tr, status )
    IF_NOTOK_RETURN(status=1)
      
    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_TQ



  ! ***



  subroutine Meteo_SetupMass( n, status )
  
    use global_data, only : mass_dat
    use dims       , only : newsrun
    use dims       , only : xcyc, im, jm
    use geometry   , only : geomtryv
    
    ! --- in/out -----------------------------
    
    integer, intent(in)       ::  n           ! region
    integer, intent(out)      ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_SetupMass'
    
    ! --- begin ------------------------------
    
    call goLabel(rname)

    ! compute initial pressure levels and mass ?
    if ( newsrun ) then
      call geomtryv( n )
    end if

    ! periodic boundary for m
    ! NOTE: m has been advected or created by geomtryv
    if ( xcyc(n) == 1 ) then  
      mass_dat(n)%m_t(0      ,:,:) = mass_dat(n)%m_t(im(n),:,:)
      mass_dat(n)%m_t(im(n)+1,:,:) = mass_dat(n)%m_t(1    ,:,:)
    end if

    ! ok
    status = 0
    call goLabel()
    
  end subroutine Meteo_SetupMass
  

  ! ***


  subroutine Meteo_CheckPressure( n, status )
  
    use ParTools   , only : myid, root, Par_AllReduce, ntracetloc
    use dims       , only : idate, newsrun
    use dims       , only : xcyc, im, jm
    use redgridZoom, only : calc_pdiff
    use io_hdf     , only : io_write2d_32d, DFACC_CREATE

    ! --- in/out -----------------------------
    
    integer, intent(in)       ::  n              ! region
    integer, intent(out)      ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Meteo_CheckPressure'
    
    ! maximum accepted pressure difference:
    real, parameter  ::  pdiffmax_treshold = 1.0e2   ! Pa
    
    ! --- external -------------------------
    
    integer(4), external    ::  sfStart, sfEnd
    
    ! --- local -----------------------------
    
    real                    ::  pdiffmax, pdiffmax_l
    integer(4)              ::  io

    ! --- begin ------------------------------
    
    call goLabel(rname)
    
    ! compare 'advected' pressure with read pressure
    if ( .not. newsrun ) then

      ! compute difference between 'advected' pressure sp and read pressure sp1;
      ! if no tracer have to be transported on this pe, use dummy:
      if ( ntracetloc == 0 ) then
        pdiffmax_l = 0.0
      else
        call calc_pdiff( n, sp1_dat(n)%data(:,:,1), &
                             sp_dat(n)%data(:,:,1), pdiffmax_l )
      end if

      ! compute maximum over all pe's:
      call Par_AllReduce( pdiffmax_l, pdiffmax, 'max', status )
      IF_NOTOK_RETURN(status=1)

      ! check ...
      if ( pdiffmax > pdiffmax_treshold ) then
        write (gol,'("difference between advected and read-in pressure exceeds treshold :")'); call goErr
        write (gol,'("  max diff. : ",es8.2," [Pa]")') pdiffmax; call goErr
        write (gol,'("  treshold  : ",es8.2," [Pa]")') pdiffmax_treshold; call goErr
        write (gol,'("pressure arrays saved to local `pressure.hdf`")'); call goErr
        if ( myid == root ) then
          io = sfStart( 'pressure.hdf', DFACC_CREATE )
          if ( io > 0 ) then
            call io_write2d_32d( io, im(n)+4, 'LON', jm(n)+4, 'LAT', sp1_dat(n)%data(:,:,1), 'p'   , idate )
            call io_write2d_32d( io, im(n)+4, 'LON', jm(n)+4, 'LAT', sp_dat(n)%data(:,:,1), 'pold', idate )
            status = sfend(io)
          else
            write (gol,'("writing pressures")'); call goErr
          end if
        end if   ! root
        call goErr; status=1; return
      end if   ! max diff

    end if  ! no newsrun

    ! ok
    status = 0
    call goLabel()
    
  end subroutine Meteo_CheckPressure



  ! **************************************************************
  ! ***
  ! *** vertical velocity
  ! ***
  ! **************************************************************
  
  subroutine Compute_Omega( omega, lli, mfw, status )
  
    use binas    , only : grav
    use grid     , only : TllGridInfo, AreaOper
    use meteodata, only : TMeteoData
    use tmm      , only : SetHistory, AddHistory
  
    ! --- in/out ----------------------------------
    
    type(TMeteoData), intent(inout)       ::  omega          ! Pa/s downward
    type(TllGridInfo), intent(in)         ::  lli
    type(TMeteoData), intent(in)          ::  mfw            ! kg/s upward
    integer, intent(out)                  ::  status
    
    ! --- const -----------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Compute_Omega'
    
    ! --- local ----------------------------------
    
    integer         ::  l
    
    ! --- begin ----------------------------------
    
    ! not in use ?    
    if ( .not. omega%used ) return

    call goLabel(rname)

    ! Pa/s  =  kg/s / m2 * g
    
    ! init with mass flux; revert sign from upward to downard, devide by gravity accelaration:
    omega%data = - mfw%data * grav   !  Pa/s m2

    ! loop over levels:
    do l = 1, size(omega%data,3)
      ! devide by cell area (m2) :
      call AreaOper( lli, omega%data(:,:,l), '/', 'm2', status )
      IF_NOTOK_RETURN(status=1)
    end do
    
    ! info ..
    !call SetHistory( omega%tmi, mfw%tmi, status )
    !call AddHistory( omega%tmi, 'convert to Pa/s', status )

    ! ok
    status = 0
    call goLabel()
    
  end subroutine Compute_Omega
  

  ! **************************************************************
  ! ***
  ! *** convective fluxes
  ! ***
  ! **************************************************************
  

  subroutine Setup_Convec( entu, entd, detu, detd, omega, &
                                tr, lli, levi, status )

    use GO       , only : TDate, wrtgol, operator(/=)
    use Grid     , only : TllGridInfo, TLevelInfo
    use TMM      , only : TMeteoInfo, Read_Convec, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  entu, entd, detu, detd
    type(TMeteoData), intent(in)          ::  omega
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    type(TLevelInfo), intent(in)          ::  levi
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_Convec'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2

    real, allocatable               ::  tmp_sp(:,:)

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! leave if not in use:
    if ( (.not. all((/entu%used,entd%used,detu%used,detd%used/)) ) &
         .and.  any((/entu%used,entd%used,detu%used,detd%used/))   ) then
      write (gol,'("either none or all of entu/entd/detu/detd should be in use")'); call goErr
      call goErr; status=1; return
    end if
    if ( .not. entu%used ) then
      call goLabel(); status=0; return
    end if
    
    ! omega is required as input:
    if ( .not. omega%used ) then
      write (gol,'("omega should be in use to compute convective stuff")'); call goErr
      call goErr; status=1; return
    end if
    
    ! not changed by default
    entu%changed = .false.
    entd%changed = .false.
    detu%changed = .false.
    detd%changed = .false.

    !
    ! time stuff
    !

    ! (sufficient to setup from entu only)
    call SetupSetup( entu, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
           
    !
    ! read/write field
    !

    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data1_t2 /= data1_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data1_t1  : ', data1_t1  ); call goErr
          call wrtgol( '  data1_t2  : ', data1_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(entu%is(1):entu%is(2),entu%js(1):entu%js(2)) )

        ! fill data:
        call Read_Convec( tmmd, entu%sourcekey, &
                 data1_tref, data1_t1, data1_t2, lli, levi, &
                 omega%data, omega%tmi, &
                 tmp_sp, &
                 entu%data1, entu%tmi1, entd%data1, entd%tmi1, &
                 detu%data1, detu%tmi1, detd%data1, detd%tmi1, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( entu%putout ) then
          call WriteField( tmmd, entu%destkey, &
                   entu%tmi1, 'sp', trim(entu%name), trim(entu%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, entu%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( entd%putout ) then
          call WriteField( tmmd, entd%destkey, &
                   entd%tmi1, 'sp', trim(entd%name), trim(entd%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, entd%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( detu%putout ) then
          call WriteField( tmmd, detu%destkey, &
                   detu%tmi1, 'sp', trim(detu%name), trim(detu%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, detu%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( detd%putout ) then
          call WriteField( tmmd, detd%destkey, &
                   detd%tmi1, 'sp', trim(detd%name), trim(detd%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, detd%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )
      
      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( entu%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( entd%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( detu%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( detd%data1, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
      entu%filled1  = .true.
      entu%tr1(1)   = data1_t1
      entu%tr1(2)   = data1_t2
      entu%changed  = .true.
      entd%filled1  = .true.
      entd%tr1(1)   = data1_t1
      entd%tr1(2)   = data1_t2
      entd%changed  = .true.
      detu%filled1  = .true.
      detu%tr1(1)   = data1_t1
      detu%tr1(2)   = data1_t2
      detu%changed  = .true.
      detd%filled1  = .true.
      detd%tr1(1)   = data1_t1
      detd%tr1(2)   = data1_t2
      detd%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
      entu%data1 = entu%data2
      entd%data1 = entd%data2
      detu%data1 = detu%data2
      detd%data1 = detd%data2
      
      ! data array is filled now:
      entu%filled1  = .true.
      entu%tr1(1)   = data1_t1
      entu%tr1(2)   = data1_t2
      entu%changed  = .true.
      entd%filled1  = .true.
      entd%tr1(1)   = data1_t1
      entd%tr1(2)   = data1_t2
      entd%changed  = .true.
      detu%filled1  = .true.
      detu%tr1(1)   = data1_t1
      detu%tr1(2)   = data1_t2
      detu%changed  = .true.
      detd%filled1  = .true.
      detd%tr1(1)   = data1_t1
      detd%tr1(2)   = data1_t2
      detd%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data2_t2 /= data2_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data2_t1  : ', data2_t1  ); call goErr
          call wrtgol( '  data2_t2  : ', data2_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(entu%is(1):entu%is(2),entu%js(1):entu%js(2)) )

        ! fill data2:
        call Read_Convec( tmmd, entu%sourcekey, &
                 data2_tref, data2_t1, data2_t2, lli, levi, &
                 omega%data, omega%tmi, &
                 tmp_sp, &
                 entu%data2, entu%tmi2, entd%data2, entd%tmi2, &
                 detu%data2, detu%tmi2, detd%data2, detd%tmi2, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( entu%putout ) then
          call WriteField( tmmd, entu%destkey, &
                   entu%tmi2, 'sp', trim(entu%name), trim(entu%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, entu%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( entd%putout ) then
          call WriteField( tmmd, entd%destkey, &
                   entd%tmi2, 'sp', trim(entd%name), trim(entd%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, entd%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( detu%putout ) then
          call WriteField( tmmd, detu%destkey, &
                   detu%tmi2, 'sp', trim(detu%name), trim(detu%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, detu%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( detd%putout ) then
          call WriteField( tmmd, detd%destkey, &
                   detd%tmi2, 'sp', trim(detd%name), trim(detd%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'n', levi, '*', &
                   tmp_sp, detd%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )
      
      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( entu%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( entd%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( detu%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( detd%data2, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data2 array is filled now:
      entu%filled2  = .true.
      entu%tr2(1)   = data2_t1
      entu%tr2(2)   = data2_t2
      entd%filled2  = .true.
      entd%tr2(1)   = data2_t1
      entd%tr2(2)   = data2_t2
      detu%filled2  = .true.
      detu%tr2(1)   = data2_t1
      detu%tr2(2)   = data2_t2
      detd%filled2  = .true.
      detd%tr2(1)   = data2_t1
      detd%tr2(2)   = data2_t2

    else if ( data2_copy ) then
    
      ! copy data2 from primary array:
      entu%data2 = entu%data1
      entd%data2 = entd%data1
      detu%data2 = detu%data1
      detd%data2 = detd%data1
      
      ! data2 array is filled now:
      entu%filled2  = .true.
      entu%tr2(1)   = data2_t1
      entu%tr2(2)   = data2_t2
      entd%filled2  = .true.
      entd%tr2(1)   = data2_t1
      entd%tr2(2)   = data2_t2
      detu%filled2  = .true.
      detu%tr2(1)   = data2_t1
      detu%tr2(2)   = data2_t2
      detd%filled2  = .true.
      detd%tr2(1)   = data2_t1
      detd%tr2(2)   = data2_t2

    end if

    
    !
    ! time interpolation
    !
    
    ! apply time interpolation:
    call TimeInterpolation( entu, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( entd, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( detu, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( detd, tr, status )
    IF_NOTOK_RETURN(status=1)
    
    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_Convec



  ! **************************************************************
  ! ***
  ! *** cloud cover
  ! ***
  ! **************************************************************
  

  subroutine Setup_CloudCovers( cc, cco, ccu, tr, lli, levi, status )

    use GO       , only : TDate, wrtgol, operator(/=)
    use Grid     , only : TllGridInfo, TLevelInfo
    use TMM      , only : TMeteoInfo, Read_CloudCovers, WriteField
    use meteodata, only : TMeteoData, TimeInterpolation
    use ParTools , only : myid, root, Par_Broadcast

    ! --- in/out ----------------------------------

    type(TMeteoData), intent(inout)       ::  cc, cco, ccu
    type(TDate), intent(in)               ::  tr(2)
    type(TllGridInfo), intent(in)         ::  lli
    type(TLevelInfo), intent(in)          ::  levi
    integer, intent(out)                  ::  status

    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Setup_Convec'
    
    ! --- local ----------------------------------
    
    logical                         ::  data1_read, data1_copy
    type(TDate)                     ::  data1_tref, data1_t1, data1_t2
    logical                         ::  data2_read, data2_copy
    type(TDate)                     ::  data2_tref, data2_t1, data2_t2

    real, allocatable               ::  tmp_sp(:,:)

    ! --- begin -----------------------------
    
    call goLabel(rname)
    
    ! leave if not in use:
    if ( (.not. all((/cc%used,cco%used/)) ) .and. any((/cc%used,cco%used/)) ) then
      write (gol,'("either none or all of entu/entd/detu/detd should be in use")'); call goErr
      call goErr; status=1; return
    end if
    if ( .not. cc%used ) then
      call goLabel(); status=0; return
    end if
    
    ! not changed by default
     cc%changed = .false.
    cco%changed = .false.
    ccu%changed = .false.

    !
    ! time stuff
    !

    ! (sufficient to setup from cc only)
    call SetupSetup( cc, tr, &
                      data1_read, data1_copy, data1_tref, data1_t1, data1_t2, &
                      data2_read, data2_copy, data2_tref, data2_t1, data2_t2, &
                      status )
    IF_NOTOK_RETURN(status=1)
           
    !
    ! read/write field
    !

    ! primairy field ?
    if ( data1_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data1_t2 /= data1_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data1_t1  : ', data1_t1  ); call goErr
          call wrtgol( '  data1_t2  : ', data1_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(cc%is(1):cc%is(2),cc%js(1):cc%js(2)) )

        ! fill data:
        call Read_CloudCovers( tmmd, cc%sourcekey, &
                 data1_tref, data1_t1, data1_t2, lli, levi, &
                 tmp_sp, cc%data1, cc%tmi1, &
                    cco%data1, cco%tmi1, ccu%data1, ccu%tmi1, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if ( cc%putout ) then
          call WriteField( tmmd, cc%destkey, &
                   cc%tmi1, 'sp', trim(cc%name), trim(cc%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, cc%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( cco%putout ) then
          call WriteField( tmmd, cco%destkey, &
                   cco%tmi1, 'sp', trim(cco%name), trim(cco%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, cco%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( ccu%putout ) then
          call WriteField( tmmd, ccu%destkey, &
                   ccu%tmi1, 'sp', trim(ccu%name), trim(ccu%unit), &
                   data1_tref, data1_t1, data1_t2, &
                   lli, 'n', levi, 'n', &
                   tmp_sp, ccu%data1, status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )
      
      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast( cc%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( cco%data1, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( ccu%data1, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data array is filled now:
       cc%filled1  = .true.
       cc%tr1(1)   = data1_t1
       cc%tr1(2)   = data1_t2
       cc%changed  = .true.
      cco%filled1  = .true.
      cco%tr1(1)   = data1_t1
      cco%tr1(2)   = data1_t2
      cco%changed  = .true.
      ccu%filled1  = .true.
      ccu%tr1(1)   = data1_t1
      ccu%tr1(2)   = data1_t2
      ccu%changed  = .true.

    else if ( data1_copy ) then
    
      ! copy data from secondary array:
       cc%data1 =  cc%data2
      cco%data1 = cco%data2
      ccu%data1 = ccu%data2
      
      ! data array is filled now:
       cc%filled1  = .true.
       cc%tr1(1)   = data1_t1
       cc%tr1(2)   = data1_t2
       cc%changed  = .true.
      cco%filled1  = .true.
      cco%tr1(1)   = data1_t1
      cco%tr1(2)   = data1_t2
      cco%changed  = .true.
      ccu%filled1  = .true.
      ccu%tr1(1)   = data1_t1
      ccu%tr1(2)   = data1_t2
      ccu%changed  = .true.

    end if

    ! secondary field ?
    if ( data2_read ) then

      if (myid==root) then ! only root does IO
      
        ! safety check ...
        if ( data2_t2 /= data2_t1 ) then
          write (gol,'("not sure that this routine is correct for time intervals:")'); call goErr
          call wrtgol( '  data2_t1  : ', data2_t1  ); call goErr
          call wrtgol( '  data2_t2  : ', data2_t2  ); call goErr
          write (gol,'("please deceide what to do with surface pressures ... ")'); call goErr
          call goErr; status=1; return
        end if
          
        ! surface pressure field:
        allocate( tmp_sp(cc%is(1):cc%is(2),cc%js(1):cc%js(2)) )

        ! fill data2:
        call Read_CloudCovers( tmmd,  cc%sourcekey, &
                 data2_tref, data2_t1, data2_t2, lli, levi, &
                 tmp_sp, cc%data2,  cc%tmi2, &
                    cco%data2, cco%tmi2, ccu%data2, ccu%tmi2, &
                 status )
        IF_NOTOK_RETURN(status=1)

        ! write meteofiles ?
        if (  cc%putout ) then
          call WriteField( tmmd,  cc%destkey, &
                    cc%tmi2, 'sp', trim( cc%name), trim( cc%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'u', levi, 'n', &
                   tmp_sp,  cc%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( cco%putout ) then
          call WriteField( tmmd, cco%destkey, &
                   cco%tmi2, 'sp', trim(cco%name), trim(cco%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'u', levi, 'n', &
                   tmp_sp, cco%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        if ( ccu%putout ) then
          call WriteField( tmmd, ccu%destkey, &
                   ccu%tmi2, 'sp', trim(ccu%name), trim(ccu%unit), &
                   data2_tref, data2_t1, data2_t2, &
                   lli, 'u', levi, 'n', &
                   tmp_sp, ccu%data2, status )
          IF_NOTOK_RETURN(status=1)
        end if
        
        ! clear
        deallocate( tmp_sp )
      
      end if  ! root ?

      ! send to other processors if necessary:
      call Par_Broadcast(  cc%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( cco%data2, root, status )
      IF_NOTOK_RETURN(status=1)
      call Par_Broadcast( ccu%data2, root, status )
      IF_NOTOK_RETURN(status=1)
    
      ! data2 array is filled now:
       cc%filled2  = .true.
       cc%tr2(1)   = data2_t1
       cc%tr2(2)   = data2_t2
      cco%filled2  = .true.
      cco%tr2(1)   = data2_t1
      cco%tr2(2)   = data2_t2
      ccu%filled2  = .true.
      ccu%tr2(1)   = data2_t1
      ccu%tr2(2)   = data2_t2

    else if ( data2_copy ) then
    
      ! copy data2 from primary array:
       cc%data2 =  cc%data1
      cco%data2 = cco%data1
      ccu%data2 = ccu%data1
      
      ! data2 array is filled now:
       cc%filled2  = .true.
       cc%tr2(1)   = data2_t1
       cc%tr2(2)   = data2_t2
      cco%filled2  = .true.
      cco%tr2(1)   = data2_t1
      cco%tr2(2)   = data2_t2
      ccu%filled2  = .true.
      ccu%tr2(1)   = data2_t1
      ccu%tr2(2)   = data2_t2

    end if

    
    !
    ! time interpolation
    !
    
    ! apply time interpolation:
    call TimeInterpolation( cc, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( cco, tr, status )
    IF_NOTOK_RETURN(status=1)
    call TimeInterpolation( ccu, tr, status )
    IF_NOTOK_RETURN(status=1)

    !
    ! done
    !

    ! ok
    status = 0
    call goLabel()

  end subroutine Setup_CloudCovers



  ! **************************************************************
  ! ***
  ! *** surface fields
  ! ***
  ! **************************************************************
  

  subroutine Coarsen_emis( sfc, emis, combkey, status )
  
    use Grid        , only : FillGrid
    use global_types, only : emis_data
    use meteodata   , only : TMeteoData
    
    ! --- in/out ----------------------------------
    
    type(TMeteoData), intent(in)    ::  sfc
    type(emis_data)                 ::  emis(nregions)
    character(len=*), intent(in)    ::  combkey
    integer, intent(out)            ::  status
    
    ! --- const --------------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/Coarsen_emis'
    
    ! --- local ----------------------------------
    
    integer        ::   n
    
    ! --- begin ----------------------------------
    
    call goLabel(rname)
    
    ! coarsen ?
    if ( sfc%changed ) then
      ! loop over regions:
      do n = 1, nregions
        ! coarsen from 1x1 to emis grid:
        call FillGrid( lli(n), 'n', emis(n)%surf, &
                       lli_1x1, 'n', sfc%data(:,:,1), &
                       combkey, status )
        IF_NOTOK_RETURN(status=1)
      end do
    end if
    
    ! ok
    call goLabel(); status=0
  
  end subroutine Coarsen_emis


  ! ***


  subroutine compute_gph( region, status )

    use Dims,        only : cdebug, kdebug, okdebug, kmain
    use Dims,        only : idate, itau, revert, newsrun
    use Dims,        only : im, jm, lm
    use Dims,        only : at, bt
    use binas,       only : grav
    use global_data, only : mass_dat
    use datetime,    only : tstamp
    use ParTools,    only : myid, root

    ! --- in/out ----------------------------------

    integer, intent(in)                   ::  region
    integer, intent(out)                  ::  status

    ! --- const ----------------------------------
    
    character(len=*), parameter  ::  rname = mname//'/compute_gph'
    
    ! --- local ----------------------------------
    
    real,dimension(:,:,:),pointer        :: gph, t, q
    real,dimension(:,:,:),pointer        :: ps

    integer                              :: i,j,l
    real                                 :: tv,pdown,pup

    ! --- begin -----------------------------
    
    ! leave if not in use:
    if ( .not. gph_dat(region)%used ) then
      !write (*,'(" gph_dat not in use")')
      status=0; return
    end if
    
    ! other meteo required:
    if ( (.not. temper_dat(region)%used) .or. (.not. humid_dat(region)%used) ) then
      write (gol,'("computation of gph requires temper and humid")'); call goErr
      write (gol,'("in ",a)') rname; call goErr; status=1; return
    end if
    
    ! leave if input did not change:
    if ( (.not.     sp_dat(region)%changed) .and. &
         (.not. temper_dat(region)%changed) .and. &
         (.not.  humid_dat(region)%changed) ) then
      !write (gol,'(a,": not changed")') rname; call goErr
      return
    end if
    
    ! field will be changed ...
    gph_dat(region)%changed = .true.

    ! pointers to meteo field
    ps  =>     sp_dat(region)%data
    t   => temper_dat(region)%data
    q   =>  humid_dat(region)%data
    gph =>    gph_dat(region)%data
    
    if ( myid==root .and. cdebug  ) call tstamp(kdebug,itau,rname)
    if ( myid==root .and. okdebug ) call tstamp(kmain ,itau,rname)

    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! compute geo potential height 
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    gph(:,:,1)=oro(region)%surf/grav   !oro is stored in g*m
    do l=1,lm(region)-1
       do j=1,jm(region)
          do i=1,im(region)
             tv = t(i,j,l)*(1. + 0.608*q(i,j,l))
             pdown = at(l) + bt(l)*ps(i,j,1)
             pup   = at(l+1) + bt(l+1)*ps(i,j,1)
             ! rgas in different units!
             gph(i,j,l+1) = gph(i,j,l) + tv*287.05*alog(pdown/pup)/grav
             ! note dec 2002 (MK) gph now from 1--->lm+1
          end do
       end do
    end do
    !set top of atmosphere at 200 km !note_mk: now in lm+1
    gph(:,:,lm(region)+1) = 200000.0

    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    ! done
    ! ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    ! done
    nullify( ps )
    nullify( t )
    nullify( q )
    nullify( gph )
    
    ! ok
    status = 0

  end subroutine compute_gph


end module meteo
