!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
module user_output_flux1x1

  !
  ! handles initialisation, accumulation and output of mean mixing ratio data
  !

  use dims
  use emission_common,    only: declare_emission_averagers, free_emission_averagers, calc_emission_averagers, save_emission_averagers, reset_emission_averagers
  use GO,                  only: gol, goErr, goPr, goLabel

  implicit none

  private

  public :: init_flux1x1
  public :: free_flux1x1
  public :: accumulate_flux1x1 
  public :: reset_flux1x1
  public :: write_flux1x1

  character(len=*), parameter  ::  mname = 'user_output_flux1x1'
  character(len=*), parameter  ::  rname = mname//'/subroutine...'

contains



!===========================================================================================================
!===========================================================================================================


  subroutine init_flux1x1

    implicit none

    !__IO____________________________________________________________________

    !__LOCAL_VARIABLES_______________________________________________________

    !__START_SUBROUTINE______________________________________________________

    ! nothing to be done

  end subroutine init_flux1x1

!===========================================================================================================
!===========================================================================================================


  subroutine reset_flux1x1

    implicit none

    !__IO____________________________________________________________________


    !__LOCAL_VARIABLES_______________________________________________________

    !__START_SUBROUTINE______________________________________________________

    call reset_emission_averagers

  end subroutine reset_flux1x1


!===========================================================================================================
!===========================================================================================================

  subroutine free_flux1x1
    implicit none

    !__IO____________________________________________________________________

    !__LOCAL_VARIABLES_______________________________________________________

    integer :: region

    !__START_SUBROUTINE______________________________________________________

    call free_emission_averagers

  end subroutine free_flux1x1

!===========================================================================================================
!===========================================================================================================


  subroutine accumulate_flux1x1
    !
    !
    implicit none

    ! in/out
    ! start

    ! nothing to be done, accumulation is in emission_co2_all.F90
  end subroutine accumulate_flux1x1

  subroutine write_flux1x1(flux1x1_data_dhour)
    !------------------------------------------------------------------
    !------------------------------------------------------------------
    use dims
    use global_data, only : outdir
    use file_hdf,    only: THdfFile, TSds, Init, Done, WriteAttribute, WriteData, SetDim
    use TMKit,       only : TDate, NewDate, IncrDate, operator(-)
    use ParTools, only : myid, root_t, Par_Barrier


    implicit none

    !__IO____________________________________________________________________
    INTEGER,INTENT(in)  :: flux1x1_data_dhour

    !__LOCAL_VARIABLES_______________________________________________________

    character(len=200)   :: FFilename

    type(THdfFile)            ::  flux1x1_hdf
    type(TDate)               ::  t1, t2
    integer,dimension(6)      ::  idate_write

    integer :: status

    !__START_SUBROUTINE______________________________________________________

    call calc_emission_averagers

    ! set interval averaging times 
    t2 = NewDate( time6=idate )
    t1 = t2 - IncrDate( hour=flux1x1_data_dhour )

    if ( myid == root_t ) then
      write(FFilename,'(a,a9,i4.4,3(i2.2),a4)') &
        trim(outdir), &
        '/flux1x1_', &
        t1%year, &
        t1%month, &
        t1%day, &
        t1%hour, &
        '.hdf'

      call Init( flux1x1_hdf, trim(FFilename), 'create', status )
      IF_ERROR_RETURN(status=1)
      call WriteAttribute( flux1x1_hdf, 'interval_start'     , (/t1%year,t1%month,t1%day,t1%hour,t1%min,t1%sec/)             , status )
      IF_ERROR_RETURN(status=1)
      call WriteAttribute( flux1x1_hdf, 'interval_end'     , (/t2%year,t2%month,t2%day,t2%hour,t2%min,t2%sec/)             , status )
      IF_ERROR_RETURN(status=1)

      call save_emission_averagers(flux1x1_hdf)
      call Done( flux1x1_hdf, status )
      IF_NOTOK_RETURN(status=1)

      

    endif

    status=0

  end subroutine write_flux1x1

end module user_output_flux1x1
